"""
    This contains all helper function for lidar to image projection
"""
from typing import List, Dict

import numpy as np

from scipy.spatial.transform import Rotation as R
from stardust.components import Camera


def get_extrin(param):
    """
    get transform matrix
    """
    cam2lidar = np.identity(4)
    rotation = list(param['heading'])
    rotation = R.from_quat(rotation).as_matrix()
    translation = np.array(list(param['position']))
    cam2lidar[:3, :3] = rotation
    cam2lidar[:3, 3] = translation

    lidar2cam = np.linalg.inv(cam2lidar)

    return lidar2cam


def get_intrin(param):
    intrin = np.array([[param['intrinsic'][0], 0.0, param['intrinsic'][2], 0.0],
                       [0.0, param['intrinsic'][1], param['intrinsic'][3], 0.0],
                       [0.0, 0.0, 1.0, 0.0]])
    return intrin


def get_radial(param):
    return np.array([v for v in param['radial']])


def get_cam_type(param):
    return param['camera_type']


def decode_cam_sources(frame_info: Dict) -> List:
    """
    Decode extrin, intrin, img width and height from frames info

    Args:
        frame_info: Dict

    Return:
        param_info: List 
    """
    imgs_info = frame_info.image
    param_info = []
    for i, img_info in enumerate(imgs_info):
        cam_param = img_info.camera_param
        heading = cam_param.heading
        position = cam_param.position
        intrin = cam_param.intrinsic
        radial = cam_param.radial
        tangential = cam_param.tangential
        cam_type = cam_param.camera_type
        cam = Camera(camera_type=cam_type,
                     heading=heading,
                     position=position,
                     intrinsic=intrin,
                     radial=radial,
                     tangential=tangential)
        param_dict = dict()
        param_dict['img_width'] = img_info.width
        param_dict['img_height'] = img_info.height
        param_dict['camera'] = cam
        param_info.append(param_dict)
    return param_info


from typing import Dict
from .base_annotation import AnnotationObject


class Text(AnnotationObject):
    def __init__(self,
                 start=None,
                 end=None,
                 content=None,
                 **kwargs):
        """
        This class defines the structure of text label task

        Args:
           start:
               The start time of annotation for this text
           end:
               The end time of annotation for this text

        Returns:
           Text instance

        Examples:
            .. code-block:: python

                text = Text(start=0.57, end=12.13)
        """
        assert isinstance(start, int)
        assert isinstance(end, int)
        self.start = start
        self.end = end
        self.content = content
        AnnotationObject.__init__(self, **kwargs)

    @classmethod
    def gen_text(cls, slot, children_lst: Dict = None, parent_id: str = None, label_kind: str = None):
        """
        generate the Polygon obj

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor ID

        Returns:
            a Polygon instance
        """

        children_lst = children_lst if children_lst else {}
        parent_id = parent_id if parent_id else ""
        label_kind = label_kind if label_kind else ""

        text = Text(
            start=slot['start'],
            end=slot['length'] + slot['start'],
            content=slot['text'],
            parent=parent_id,
            id=slot['id'],
            children_lst=children_lst,
            label=label_kind
        )

        return text



from typing import Union, Optional, Dict, Sequence, List
from stardust.components.annotations.point import Point
from stardust.components.annotations.base_annotation import AnnotationObject


class Box2D(AnnotationObject):
    def __init__(self,
                 p1: Union[Sequence, Point] = None,
                 p2: Union[Sequence, Point] = None,
                 center: Union[Sequence, Point] = None,
                 size: Optional[List[float]] = None,
                 source: Optional[str] = None,
                 **kwargs) -> None:
        """
        This class defines the structure of box2d label task

        Args:
            p1:
                Point obj or a Sequence obj, the point at the top left corner of a 2d box
            p2:
                Point obj or a Sequence obj, the point at the bottom right corner of a 2d box
            center:
                Center point coordinates, supporting Point objects or sequences of length 2
            size:
                Size, supports Point objects or sequences of length 2
            source:
                camera name, string, camera name

        Returns:
            Box2D instance

        Examples:
            .. code-block:: python

                bx = Box2D(center=[473.07, 395.93], size=[38.65, 28.67])
                bx.area

                >>> output
                1108.0955000000001
        """

        if p1 is not None and p2 is not None:
            self._set_val("p1", p1)
            self._set_val("p2", p2)
            self.center = [self.width / 2 + self.p1[0], self.height / 2 + self.p1[1]]
            self.size = [self.p2[0] - self.p1[0], self.p2[1] - self.p1[1]]
            # assert self.size[0] > 0 and self.size[1] > 0
        elif center is not None and size is not None:
            self._set_val("center", center)
            self._set_val("size", size)
            self.p1 = [self.center[0] - self.size[0] / 2, self.center[1] - self.size[1] / 2]
            self.p2 = [self.center[0] + self.size[0] / 2, self.center[1] + self.size[1] / 2]
        else:
            raise TypeError("missing required parameters")

        self.source = source
        AnnotationObject.__init__(self, **kwargs)

    def _set_val(self, key, value):
        if isinstance(value, Sequence):
            assert len(value) == 2
            self.__dict__[key] = list(value)
        elif isinstance(value, Point):
            self.__dict__[key] = [value.x, value.y]
        else:
            raise TypeError('Unsupported data type')

    @property
    def corners(self) -> Dict[str, Point]:
        """
        Returns:
            float: the conner points os the current box2d
        """
        return {
            'x1': self.p1[0],
            'y1': self.p1[1],
            'x2': self.p2[0],
            'y2': self.p2[1]
        }

    @property
    def x1(self) -> float:
        """
        Returns:
            float: The x-coordinate of the top left point.
        """
        return self.p1[0]

    @property
    def y1(self) -> float:
        """
        Returns:
            float: The y-coordinate of the top left point.
        """
        return self.p1[1]

    @property
    def x2(self) -> float:
        """
        Returns:
            float: The x-coordinate of the bottom right point.
        """
        return self.p2[0]

    @property
    def y2(self) -> float:
        """
        Returns:
            float: The y-coordinate of the bottom right point.
        """
        return self.p2[0]

    @property
    def height(self) -> float:
        """
        Returns:
            float: the height of the current box2d, axis y
        """
        return round(self.p2[1] - self.p1[1], 2)

    @property
    def width(self) -> float:
        """
        Returns:
            float: the width of the current box2d, axis x
        """
        return round(self.p2[0] - self.p1[0], 2)

    @property
    def area(self) -> float:
        """
        Returns:
            float: the area of the current box2d
        """
        return self.width * self.height

    @staticmethod
    def gen_box(slot, children_lst=None, parent_id=None, label_kind=None, **kwargs):
        """
        generate the Box2D obj

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor id
        Returns:
            a Box2D instance
        """
        children_lst = children_lst if children_lst else {}
        parent_id = parent_id if parent_id else ""
        label_kind = label_kind if label_kind else ""
        box2d = Box2D([round((p := slot['plane']["topLeft"])['x'], 2), round(p['y'], 2)],
                      [round((p := slot['plane']["bottomRight"])['x'], 2), round(p['y'], 2)],
                      label=label_kind,
                      id=slot['id'],
                      source=slot.get('source'),
                      children=children_lst,
                      parent=parent_id,
                      team_id=kwargs.get('team_id', None)
                      )
        return box2d



import os
import numpy as np

from stardust.utils.convert import read_rosetta


def object_class(val, label):
    for child in val.children:
        if child in label.input_lst.keys() and label.input_lst[child].name == 'category':
            box_class = label.input_lst[child].value
            if isinstance(box_class, list):
                return '/'.join(box_class[0])
            else:
                return box_class
    return 'unknown'


def camera_map(camera_param):
    return {cam.name: cam.camera_param.cam2lidar_rt.I.T for cam in camera_param}


def box_center_in_image(cam, center):
    center = np.c_[center.reshape(1, 3), np.array([1])]
    center = np.dot(center, cam).tolist()[0][:3]
    return center


def kitti_export(stardust_data: "Frame", save_path: str):
    """
    rosetta to kitti

    Args:
        stardust_data:
            Frame class
        save_path:
            Save path
    """
    while True:
        try:
            json_data = next(stardust_data)
        except AssertionError:
            pass
        except StopIteration:
            break
        else:
            try:
                camera_info = camera_map(json_data.media.image)
            except Exception as e:
                raise ValueError(f'Camera parameters are invalid: {e}')
            txt_name = os.path.basename(json_data.media.point_cloud.uri).replace('pcd', 'txt')
            os.makedirs(save_path, exist_ok=True)
            with open(os.path.join(save_path, txt_name), 'w') as wp:
                annotation = json_data.annotation
                for key, value in annotation.box2d_lst.items():
                    # category, truncation, occlusion, observation angle, 2D top-left, 2D bottom-right, 3D dimensions, 3D position, yaw, score
                    kitti_data = [None] * 16
                    kitti_data[4:6] = value.x1, value.y1
                    kitti_data[6:8] = value.x2, value.y2
                    box3d_info = annotation.box3d_lst.get(value.parent, None)
                    if box3d_info:
                        kitti_data[0] = object_class(box3d_info, annotation)
                        kitti_data[8:11] = box3d_info.size
                        kitti_data[11:14] = box_center_in_image(camera_info[value.source], np.array(box3d_info.center))
                        kitti_data[14] = box3d_info.rotation[-1]
                    else:
                        kitti_data[0] = object_class(value, annotation)
                    wp.write(' '.join(np.array(kitti_data, dtype=str).tolist()))
                    wp.write("\n")
    print(f"Exported KITTI files saved to: {save_path}")


if __name__ == '__main__':
    project_id = 1507
    input_path = f'/Users/mac/Desktop/{project_id}'

    _json_data = read_rosetta(project_id=project_id,
                              input_path=input_path,
                              )
    kitti_export(_json_data, os.path.join(input_path, 'kitti'))



"""
    Get rosetta pool information
"""
import requests

from stardust.rosetta.rosetta_token import RosettaToken
from stardust.config import RosConfig

__all__ = ['RosettaPool']


class RosettaPool:
    sever_map = {
        'top': RosConfig.env_prod,
        'dev': RosConfig.env_dev
    }

    def __init__(self, env: str = 'top'):
        self.env = self.sever_map[env]
        self.headers: dict = RosettaToken(self.env).get_headers()

    def get_pool_info(self, project_id: int):
        url = f'{self.env}/rosetta-service/project/doneTask/get?projectId={project_id}'
        res = requests.get(url, headers=self.headers)
        res_data = res.json()
        pool_list = [{
            i['poolId']: i['poolName']} for i in res_data['data']]
        return pool_list

    def get_all_finish_pool_info(self, project_id: int):
        url = f'{self.env}/rosetta-service/project/doneTask/get?projectId={project_id}'
        res = requests.get(url, headers=self.headers)
        res_data = res.json()
        assert res_data['code'] not in  (5802, )
        pool_list = [i['poolId'] for i in res_data['data']]
        assert pool_list, ValueError(f'Project {project_id} has no completed pools')
        return pool_list


if __name__ == '__main__':
    pool = RosettaPool()
    print(pool.get_pool_info(1337))
    print(pool.get_all_finish_pool_info(1337))



from typing import Union, Sequence, Dict
from scipy.spatial.transform import Rotation

import numpy as np
import open3d as o3d

from .base_annotation import AnnotationObject
from .point3d import Point3D
from stardust.utils.matrix4 import Matrix4


class Box3D(AnnotationObject):
    def __init__(self,
                 center: Union[Point3D, Sequence],
                 size: Union[Point3D, Sequence],
                 rotation: Sequence,
                 rotation_order="XYZ",
                 **kwargs
                 ) -> None:
        """
        This class defines the structure of 3d geometry box.

        Args:
            center:
                like: [4.133927784395542, -3.777526126541506, 0.7891658530222055] or Point3D obj
            size:
                like [7.298632312111001, 2.192134290714646, 1.9819562591803783], Iterable， extent in x, y and z direction or Point3D obj
            rotation:
                like [0, 0, -1.5713694508648803], Iterable
            rotation_order:
                default to be XYZ

        Returns:
            Box3D instance

        Examples:
            .. code-block:: python

                box = Box3D(Point3D(4.13, -3.77, 0.78),Point3D(1, 5, 1),[0, 0, -1.57])

                box.volume

                >>> output:
                >>> 5
        """
        # Currently only supports XYZ order
        assert len(rotation) == 3
        self._set_val("center", center)
        self._set_val("size", size)

        if volume := self._verify_is_box():
            self.volume = volume
        else:
            raise ValueError("Insufficient inputs to form a 3D box")
        self.rotation = rotation
        self.rotation_order = rotation_order

        AnnotationObject.__init__(self, **kwargs)

    def _set_val(self, key, value):
        if isinstance(value, Sequence):
            assert len(value) == 3
            self.__dict__[key] = list(value)
        elif isinstance(value, Point3D):
            self.__dict__[key] = [value.x, value.y, value.z]
        else:
            raise TypeError('Unsupported data type')

    def _verify_is_box(self) -> Union[bool, float]:
        """
        To verify if the input can form a 3D box

        Returns: bool/float
        """
        volume = self.size[0] * self.size[1] * self.size[2]
        # Ensure the oriented bounding box volume is positive
        if volume > 0:
            return volume
        else:
            return False

    @property
    def vertices(self) -> np.array:
        """
        The 8 vertices of the current 3D frame object

        Returns:
            np.array, the points from box3d

        Examples:
            .. code-block:: python

                box.vertices

                >>> output:

                >>> np.array([[ 1.63421476 -3.2760934   0.28916585]
                 [ 1.63364163 -4.27609323  0.28916585]
                 [ 6.63421394 -3.27895902  0.28916585]
                 [ 1.63421476 -3.2760934   1.28916585]
                 [ 6.63364081 -4.27895885  1.28916585]
                 [ 6.63421394 -3.27895902  1.28916585]
                 [ 1.63364163 -4.27609323  1.28916585]
                 [ 6.63364081 -4.27895885  0.28916585]])
        """
        obb = o3d.geometry.OrientedBoundingBox()
        center = self.center
        size = self.size
        rotation = self.rotation
        obb.center = np.array([center.x, center.y, center.z]) if isinstance(center, Point3D) else center
        obb.extent = np.array([size.x, size.y, size.z]) if isinstance(size, Point3D) else size
        obb.R = Rotation.from_euler(self.rotation_order, rotation).as_matrix()
        return np.array(obb.get_box_points())

    @staticmethod
    def gen_box3d(slot, children_lst: Dict = None, parent_id: str = None, label_kind: str = None, team_id: int = None):
        """
        generate the Box3D obj

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor ID

        Returns:
                a Box3D instance
        """
        if not slot.get("box", None):
            return
        box = Matrix4(slot['box']).box_info
        children_lst = children_lst if children_lst else {}
        parent_id = parent_id if parent_id else ""
        label_kind = label_kind if label_kind else ""
        box3d = Box3D(
            id=slot['id'],
            center=[(p := box['position'])['x'], p['y'], p['z']],
            size=[(s := box['scale'])['height'], s['width'], s['depth'] if s.get('depth') else s.get('length')],
            rotation=[(r := box['position'])['x'], r['y'], r['z']],
            rotation_order="XYZ",
            parent=parent_id,
            label=label_kind,
            children=children_lst,
            team_id=team_id
        )

        return box3d



"""
Utilities for Alibaba Cloud OSS storage.
"""
import oss2
from typing import List
from stardust.config import OssUser as OssConfig

__all__ = ["oss"]


# Account configuration
class OssAuth(object):
    def __init__(self):
        self.oss_key = OssConfig.oss_key
        self.oss_secret = OssConfig.oss_secret
        self.endpoint = OssConfig.endpoint
        self.prefix_url = OssConfig.prefix_url
        self.auth = oss2.Auth(self.oss_key, self.oss_secret)


class Oss(OssAuth):
    __instance = None

    def __new__(cls):
        if not cls.__instance:
            cls.__instance = super(Oss, cls).__new__(cls)
            return cls.__instance
        else:
            return cls.__instance

    def __init__(self, bucket=OssConfig.oss_bucket, connect_timeout=None, enable_crc=True):
        super().__init__()
        self.bucket = oss2.Bucket(self.auth,
                                  self.endpoint,
                                  bucket,
                                  connect_timeout=connect_timeout,
                                  enable_crc=enable_crc)

    def load_oss_file_lst(self, file_lst: List = None):
        if not file_lst:
            return "empty"
        # Build batch download request
        batch = oss2.BatchGetObjectsRequest(bucket_name)
        for obj in object_list:
            batch.add_key(obj)

        # Execute batch download request
        result = bucket.batch_get_objects(batch)

        # Process download results
        for i, obj in enumerate(result.object_list):
            if obj.status == 200:
                save_path = 'path/to/save/' + object_list[i]  # Update save path to match your environment
                with open(save_path, 'wb') as f:
                    f.write(obj.read())
            else:
                print(f"Failed to download {object_list[i]}")

        print("Download completed.")


oss = Oss()
if __name__ == '__main__':
    oss = Oss()
    print(oss.bucket.object_exists("Clients/COSMO/ChatDataset/Production/LDC/.DS_Store"))



from typing import Union, Optional, List

from stardust.components.attachment import *


class Media:
    def __init__(
            self,
            point_cloud: Optional[Union[PointCloud, PointCloudSet]] = None,
            image: Optional[Union[Image, ImageSet]] = None
    ):
        """
        Annotate container objects such as files, point clouds, pictures, audio and video

        Args:
            point_cloud:
                Point cloud, point cloud set

                eg:
                    point_cloud = PointCloud("1681532060081818.pcd")
            image:
                image、image—set，eg: image = Image("1689737321.jpg")

        Returns：
            Media instance

        Examples:
            .. code-block:: python

                media = Media(point_cloud=PointCloud(*), image=Image(*))

        """
        self.point_cloud = point_cloud
        self.image = image


class TaskInfo:
    def __init__(self,
                 task_id: int = None,
                 frame_num: int = None,
                 project_id: int = None,
                 pool_id: int = None,
                 ):
        """
        rosetta Information about the project

        Args:
            task_id:
                int, rosetta_task_id
            frame_num:
                int, frame, Start at 0
            project_id:
                int, rosetta_project_id
            pool_id:
                int, rosetta_pool_id
        Returns:
            TaskInfo instance

        Examples:
            .. code-block:: python

                task_info = TaskInfo(288919499858784002, 0, 1703, 456212)

        """
        self.task_id = task_id
        self.frame_num = frame_num
        self.project_id = project_id
        self.pool_id = pool_id


class Annotation:
    def __init__(self,
                 box3d_lst=None,
                 box2d_lst=None,
                 input_lst=None,
                 line_lst=None,
                 polygon_lst=None,
                 key_point_lst=None,
                 text_lst=None
                 ):
        """
        Annotate the result, supporting 3D box, 2D box, input box, line, polygon

        Args:
            box3d_lst:
                All box3d annotation results, in general, take the box3d object id as the key and box3d object as the value
            box2d_lst:
                All box2d annotation results, in general, take the box2d object id as the key and box2d object as the value
            input_lst:
                In general, the input object id is used as the key and the input object as the value
            line_lst:
                In general, the result of all line annotations takes the id of the line object as the key and the line object as the value
            polygon_lst:
                All polygon annotations, in general, take the id of the polygon object as the key and the polygon object as the value
        Returns:
            Annotation instance

        Examples:
            .. code-block:: python

                ann = Annotation(
                    box3d_lst,
                    box2d_lst,
                    input_lst,
                    line_lst,
                    polygon_lst,
                )

        """
        if not box3d_lst:
            box3d_lst = {}
        if not box2d_lst:
            box2d_lst = {}
        if not input_lst:
            input_lst = {}
        if not line_lst:
            line_lst = {}
        if not polygon_lst:
            polygon_lst = {}
        if not key_point_lst:
            key_point_lst = {}
        if not text_lst:
            text_lst = {}

        self.box3d_lst = box3d_lst
        self.box2d_lst = box2d_lst
        self.input_lst = input_lst
        self.line_lst = line_lst
        self.polygon_lst = polygon_lst
        self.key_point_lst = key_point_lst
        self.text_lst = text_lst


Prediction = Annotation


class Frame:
    def __init__(self,
                 media: Media = None,
                 task_info: TaskInfo = None,
                 annotation: Annotation = None,
                 prediction: Prediction = None,
                 ):
        """
        Single frame structure

        Args:
            media:
                Media obj
            task_info:
                TaskInfo obj
            annotation:
                Annotation obj

        Returns:
            Frame instance

        Examples:
            .. code-block:: python

                frame = Frame(media,
                    task_info,
                    annotation,
                    prediction
                )
        """
        self.media = media
        self.task_info = task_info
        self.annotation = annotation
        self.prediction = prediction

    @property
    def urls(self) -> List[str]:
        """
        Gets the address of the annotated file

        Returns:
            list
        """
        lst = list()
        if self.media.point_cloud.uri:
            lst = [self.media.point_cloud.uri, ]
        if isinstance(self.media.image, list):
            lst.extend([image.uri for image in self.media.image])
        return lst


if __name__ == '__main__':
    pass



import os
import json
import numpy as np
from glob import glob
from tqdm import tqdm
from typing import Union, Optional
from concurrent.futures import ProcessPoolExecutor, as_completed

from stardust.rosetta.new_structure import Slot
from stardust.rosetta.mapping_structure import SlotMapping
from stardust.components.camera import Camera
from stardust.utils.frames import *


# Definition of sequential frames
class Frames:
    sequence_type_list = ["IMAGE_SEQUENCE", "IMAGE_SET_SEQUENCE", "POINTCLOUD_SEQUENCE", "POINTCLOUD_SET_SEQUENCE"]

    @classmethod
    def read_json(cls, path: str):
        with open(path) as fp:
            cls._ = json.load(fp)

    def __init__(self, path: str):
        self.read_json(path)
        self.path = path
        record: dict = self._['taskParams']['record']
        self.attachment_type: str = record['attachmentType']
        assert self.attachment_type in self.sequence_type_list, 'Not supported'
        self.projectId: int = self._['projectId']
        self.datasetId: int = self._['datasetId']
        self.poolId: int = self._['poolId']
        self.taskId: int = self._['taskId']
        self.status: int = self._['status']
        self.attachment: list = record['attachment']
        self.attachment_length: int = len(record['attachment'])
        self.metadata: dict = record.get('metadata', {})
        self.operators: list = self._['taskParams']['operators']
        self.result_annotations: list = self._['result']['annotations']
        self.result_hints: list = self._['result'].get('hints', [])
        self.result_metadata: dict = self._['result'].get('metadata', {})
        self.error_info: str = f'project{self.projectId}_pool{self.poolId}_task{self.taskId}'

    def frame_structure(self, frame_number) -> dict:
        return {
            "projectId": self.projectId,
            "datasetId": self.datasetId,
            "poolId": self.poolId,
            "taskId": self.taskId,
            "status": self.status,
            "taskParams": {
                "record": {
                    "attachmentType": self.attachment_type.replace("_SEQUENCE", ""),
                    "attachment": self.attachment[frame_number],
                    "metadata": self.metadata
                },
                "operators": self.operators
            }
        }

    @property
    def empty_operators(self):
        return [slot_structure(_) for _ in self.operators]

    def operators_record(self, operators: Optional[list] = None, operators_record: Optional[dict] = None):
        if not operators:
            operators = self.operators
        if operators_record is None:
            operators_record = {}
        for op in operators:
            if op['type'] != 'slotChildren':
                continue
            operators_record.setdefault(op['key'], op['slotSpecification'].get('imageSourceMapping', []))
            self.operators_record(op['children'], operators_record)
        return operators_record

    def annotation_type_factory(self, annotation, frame_number, operators=None) -> callable:
        func = Factory(annotation,
                       frame_number,
                       self.attachment,
                       operators)
        return func.factory(annotation['type'])

    def _normal_frames_split(self, frame_number) -> list:
        normal_annotation_frame_data = []
        for annotation in self.result_annotations:
            normal_annotation_frame_data.append(self.annotation_type_factory(annotation, frame_number))
        return normal_annotation_frame_data

    def _reflection_frames_split(self, frame_number) -> list:
        reflection_annotation_frame_data = []
        operators = self.operators_record()
        for annotation in self.result_annotations:
            reflection_annotation_frame_data.append(self.annotation_type_factory(annotation, frame_number, operators))
        return reflection_annotation_frame_data

    def _frames_split(self, frame_number) -> list:
        if self.result_metadata:
            invalid_frame = self.result_metadata.get('invalidFrame', [])
            if frame_number in invalid_frame:
                frame_annotations = self.empty_operators
            else:
                if self.result_metadata.get('interpolateMethod', '') == 'mapping-first':
                    frame_annotations = self._reflection_frames_split(frame_number)
                else:
                    frame_annotations = self._normal_frames_split(frame_number)
        else:
            frame_annotations = self._normal_frames_split(frame_number)
        return frame_annotations

    def frames_split(self):
        for frame_number in range(self.attachment_length):
            new_path = frame_name(self.path, frame_number)
            new_path = os.path.join(os.path.dirname(new_path), str(self.taskId), os.path.basename(new_path))
            os.makedirs(os.path.dirname(new_path), exist_ok=True)
            with open(new_path, "w") as fw:
                frame_data = self.frame_structure(frame_number)
                try:
                    frame_annotations = self._frames_split(frame_number)
                except Exception as e:
                    print(f'{self.error_info}_frame{frame_number} split failed: {e}')
                    frame_annotations = self.empty_operators
                frame_data['result'] = {
                    'annotations': frame_annotations,
                    'hints': self.result_hints,
                    'metadata': self.result_metadata
                }
                fw.write(json.dumps(frame_data, ensure_ascii=False))
        os.remove(self.path)


class Factory:
    def __init__(
            self,
            annotation: dict,
            frame_number: int,
            attachment: list,
            operators: Optional[list],
            reflection: bool = False
    ):
        self.annotation = annotation
        self.frame_number = frame_number
        self.attachment = attachment
        self.operators = operators
        self.reflection = reflection

    def image_source(self):
        now_attachment = self.attachment[self.frame_number].get('imageSources', [])
        assert now_attachment, 'now_attachment must not be empty'
        return {
            n['name']: {
                'camera': Camera(
                    camera_type=n['camera']['type'],
                    heading=tuple(n['camera']['heading'].values()),
                    position=tuple(n['camera']['position'].values()),
                    intrinsic=tuple(n['camera']['intrinsic'].values()),
                    radial=tuple(n['camera']['radial'].values()),
                    tangential=tuple(n['camera']['tangential'].values())
                ),
                'width': n['width'],
                'height': n['height']
            } for n in now_attachment
        }

    def annotation_slots_children(self, annotation: Optional[dict] = None):
        if not annotation:
            annotation = self.annotation
        slot_children_dict = slot_structure(annotation)
        for slot_children in annotation.get('slotsChildren', []):
            slot_ = self.annotation_slot(slot_children['slot'])
            children_ = []
            children_record = []
            for children in slot_children['children']:
                if self.operators and children['key'] in self.operators[annotation['key']]:
                    children_result = self.factory(children['type'], children, slot_)
                else:
                    children_result = self.factory(children['type'], children)
                children_record.append(empty(children_result))
                children_.append(children_result)
            if isinstance(slot_.label, bool) and np.all(np.array(children_record)):
                continue
            slot_children_dict['slotsChildren'].append({
                'slot': vars(slot_),
                'children': children_
            })
        return slot_children_dict

    def annotation_children_only(self, annotation: Optional[dict] = None):
        if not annotation:
            annotation = self.annotation
        children_only_dict = slot_structure(annotation)
        for children_only in annotation['childrenOnly']:
            children_ = []
            for children in children_only['children']:
                children_.append(self.factory(children['type'], children))
            else:
                children_only_dict['childrenOnly'].append({
                    'id': children_only['id'],
                    'label': children_only['label'],
                    'teamId': children_only.get('teamId', None),
                    'children': children_})
        return children_only_dict

    def annotation_slot(self, annotation: Optional[dict] = None, father: object = None):
        if not annotation:
            annotation = self.annotation
        if annotation['type'] == 'slot':
            slot_dict = slot_structure(annotation)
            for slot_obj in annotation['slots']:
                if father:
                    slot = SlotMapping(slot_obj, father, self.frame_number, self.image_source())
                else:
                    slot = Slot(slot_obj, self.frame_number, self.attachment)
                if isinstance(slot.label, bool):
                    continue
                slot_dict['slots'].append(vars(slot))
            return slot_dict
        else:
            attribute_slot = Slot(annotation, self.frame_number, self.attachment)
            return attribute_slot

    def annotation_input(self, annotation: Optional[dict] = None):
        if not annotation:
            annotation = self.annotation
        input_dict = slot_structure(annotation)
        try:
            input_dict['input'].update(vars(Slot(annotation, self.frame_number))['input'])
        except TypeError:
            return input_dict
        return input_dict

    def factory(self, annotation_type, annotation: dict = None, slot: object = None) -> callable:
        if not annotation:
            annotation = self.annotation
        if annotation_type == 'slotChildren':
            return self.annotation_slots_children(annotation)
        if annotation_type == 'childrenOnly':
            return self.annotation_children_only(annotation)
        if annotation_type == 'slot':
            return self.annotation_slot(annotation, slot)
        else:
            return self.annotation_input(annotation)


def _to_split(path):
    try:
        frames = Frames(path)
    except AssertionError:
        os.rename(path, frame_name(path))
    else:
        frames.frames_split()
    return True


def to_split(input_path):
    file_paths = [file for file in glob(f'{input_path}/**', recursive=True) if file.endswith('.json')]
    bar = tqdm(total=len(file_paths))
    with ProcessPoolExecutor(max_workers=4) as executor:
        futures = [executor.submit(_to_split, file_path) for file_path in file_paths]
        for future in as_completed(futures):
            if future.result():
                bar.update(1)


if __name__ == '__main__':
    import open3d as o3d
    paths = '/Users/stardust/Downloads/2073'

    # paths = glob('/Users/stardust/Downloads/1953/*.json')
    # frames_ = Frames(paths[0])

    import time

    s = time.time()
    # frames_.frames_split()
    to_split("/Users/mac/Documents/pyproject/sd_sdk/stardust_sdk/Stardust_SDK/data/2201/2201")
    # o = frames.operators_record()
    # print(o)1
    end = time.time()
    print(end - s)



from copy import deepcopy
import numpy as np
from stardust.rosetta.frame_tools import UtilityCategory, FrameCapture
from stardust.utils.matrix4 import Matrix4


class Slot(UtilityCategory):
    """slot"""
    attr_map = {
        'box3d': 'matrix',
        'polygon3d': 'points',
        'line3d': 'points',
        'point3d': 'point',
        'box2d': 'plane',
        'line': 'vertices',
        'splines': 'vertices',
        'polygon': 'vertices',
        'point': 'point',
        'cuboid': 'vertices',
        'box3d_world': 'matrix',
        'polygon3d_world': 'points',
        'line3d_world': 'points',
        'point3d_world': 'point'
    }

    def __init__(self, data, frame_number, capture):
        super().__init__(data)
        slot_map = {
            'box3d': self.box3d_time_series,
            'polygon3d': self.points_time_series,
            'line3d': self.points_time_series,
            'point3d': self.point3d_time_series,
            'box2d': self.box2d_time_series,
            'line': self.line_time_series,
            'splines': self.line_time_series,
            'polygon': self.line_time_series,
            'point': self.point_time_series,
            'cuboid': self.cuboid_time_series,
            'box3d_world': self.box3d_time_series_world,
            'polygon3d_world': self.points_time_series_world,
            'line3d_world': self.points_time_series_world,
            'point3d_world': self.point3d_time_series_world
        }
        self.id = data['id']
        self.capture = capture
        if data.get('source'):
            self.source = data['source']
        data_type = data['type']
        if data_type in ['box3d', 'polygon3d', 'line3d', 'point3d'] and data.get('coordinate') == 'world':
            slot_map[f'{data_type}_world'](self.attr_map[data_type], data['timeSeries'], frame_number)
        else:
            slot_map[data_type](self.attr_map[data_type], data['timeSeries'], frame_number)

    def box3d_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    if None in time_series_dict[str(frame_number)]['value']:
                        pass
                    else:
                        self.box = deepcopy(time_series_dict[str(frame_number)]['value'])
                        Matrix4.box = time_series_dict[str(frame_number)]['value']
                        time_series_record[frame_number] = Matrix4().box_info

            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                self.box = deepcopy(time_series_dict[closest_down]['value'])
                Matrix4.box = time_series_dict[closest_down]['value']
                time_series_record[frame_number] = Matrix4().box_info

            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                Matrix4.box = time_series_dict[closest_down]['value']
                rM = Matrix4().box_info
                Matrix4.box = time_series_dict[closest_up]['value']
                lM = Matrix4().box_info
                _frame = [closest_down, closest_up, frame_number]
                box = {
                    'position': {
                        'x': self.capture.inter_polate_number(*_frame, lM['position']['x'], rM['position']['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM['position']['y'], rM['position']['y']),
                        'z': self.capture.inter_polate_number(*_frame, lM['position']['z'], rM['position']['z'])},
                    'rotation': {
                        'x': self.capture.inter_polate_number(*_frame, lM['rotation']['x'],
                                                              self.capture.angle_closer_to(lM['rotation']['x'],
                                                                                           rM['rotation']['x'])),
                        'y': self.capture.inter_polate_number(*_frame, lM['rotation']['y'],
                                                              self.capture.angle_closer_to(lM['rotation']['y'],
                                                                                           rM['rotation']['y'])),
                        'z': self.capture.inter_polate_number(*_frame, lM['rotation']['z'],
                                                              self.capture.angle_closer_to(lM['rotation']['z'],
                                                                                           rM['rotation']['z'])),
                        '_order': 'XYZ'},
                    'scale': {
                        'height': self.capture.inter_polate_number(*_frame, lM['scale']['height'],
                                                                   rM['scale']['height']),
                        'width': self.capture.inter_polate_number(*_frame, lM['scale']['width'], rM['scale']['width']),
                        'depth': self.capture.inter_polate_number(*_frame, lM['scale']['depth'], rM['scale']['depth'])}
                }
                time_series_record[frame_number] = box
                self.box = Matrix4.get_transform_matrix(box)
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def box3d_time_series_world(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    if None in time_series_dict[str(frame_number)]['value']:
                        pass
                    else:
                        self.box = deepcopy(time_series_dict[str(frame_number)]['value'])
                        Matrix4.box = time_series_dict[str(frame_number)]['value']
                        time_series_record[frame_number] = Matrix4().box_info

            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):

                down_coordinate = self.capture.get_world(self.capture.attachment[int(closest_down)].get('coordinate'))
                frame_coordinate = self.capture.get_world(self.capture.attachment[frame_number].get('coordinate'))
                Matrix4.box = time_series_dict[closest_down]['value']
                rM = Matrix4.get_3dbox_same_frame(Matrix4().box_info, down_coordinate)
                frame = Matrix4.get_3dbox_same_frame(rM, np.linalg.inv(frame_coordinate))

                time_series_record[frame_number] = frame
                self.box = Matrix4.get_transform_matrix(frame)


            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                down_coordinate = self.capture.get_world(self.capture.attachment[int(closest_down)].get('coordinate'))
                up_coordinate = self.capture.get_world(self.capture.attachment[int(closest_up)].get('coordinate'))
                Matrix4.box = time_series_dict[closest_down]['value']
                rM = Matrix4.get_3dbox_matrix(Matrix4().box_info, down_coordinate)
                Matrix4.box = time_series_dict[closest_up]['value']
                lM = Matrix4.get_3dbox_matrix(Matrix4().box_info, up_coordinate)
                _frame = [closest_down, closest_up, frame_number]
                box = {
                    'position': {
                        'x': self.capture.inter_polate_number(*_frame, lM['position']['x'], rM['position']['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM['position']['y'], rM['position']['y']),
                        'z': self.capture.inter_polate_number(*_frame, lM['position']['z'], rM['position']['z'])},
                    'rotation': {
                        'x': self.capture.inter_polate_number(*_frame, lM['rotation']['x'],
                                                              self.capture.angle_closer_to(lM['rotation']['x'],
                                                                                           rM['rotation']['x'])),
                        'y': self.capture.inter_polate_number(*_frame, lM['rotation']['y'],
                                                              self.capture.angle_closer_to(lM['rotation']['y'],
                                                                                           rM['rotation']['y'])),
                        'z': self.capture.inter_polate_number(*_frame, lM['rotation']['z'],
                                                              self.capture.angle_closer_to(lM['rotation']['z'],
                                                                                           rM['rotation']['z'])),
                        '_order': 'XYZ'},
                    'scale': {
                        'height': self.capture.inter_polate_number(*_frame, lM['scale']['height'],
                                                                   rM['scale']['height']),
                        'width': self.capture.inter_polate_number(*_frame, lM['scale']['width'], rM['scale']['width']),
                        'depth': self.capture.inter_polate_number(*_frame, lM['scale']['depth'], rM['scale']['depth'])}
                }
                frame = Matrix4.get_3dbox_matrix(box, self.capture.get_world(
                    self.capture.attachment[frame_number].get('coordinate')).I)
                time_series_record[frame_number] = frame
                self.box = Matrix4.get_transform_matrix(frame)
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def points_time_series_world(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                frame_coordinate = self.capture.get_world(self.capture.attachment[frame_number].get('coordinate'))
                down_coordinate = self.capture.get_world(self.capture.attachment[int(closest_down)].get('coordinate'))

                points = []
                for s in time_series_dict[closest_down]['value']:
                    rM = Matrix4.get_points_matrix(s, down_coordinate)
                    points.append(Matrix4.get_points_matrix(rM, np.linalg.inv(frame_coordinate)))
                time_series_record[frame_number] = points
            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                down_coordinate = self.capture.get_world(self.capture.attachment[int(closest_down)].get('coordinate'))
                up_coordinate = self.capture.get_world(self.capture.attachment[int(closest_up)].get('coordinate'))
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                points = []
                for s in range(len(rM)):
                    rM_points = Matrix4.get_points_matrix(rM[s], down_coordinate)
                    lM_points = Matrix4.get_points_matrix(lM[s], up_coordinate)
                    frame_points = {
                        'x': self.capture.inter_polate_number(*_frame, lM_points['x'], rM_points['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM_points['y'], rM_points['y']),
                        'z': self.capture.inter_polate_number(*_frame, lM_points['z'], rM_points['z'])}
                    points.append(Matrix4.get_points_matrix(frame_points, self.capture.get_world(
                        self.capture.attachment[frame_number].get('coordinate')).I))
                time_series_record[frame_number] = points
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def point3d_time_series_world(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                frame_coordinate = self.capture.get_world(self.capture.attachment[frame_number].get('coordinate')).I
                down_coordinate = self.capture.get_world(self.capture.attachment[int(closest_down)].get('coordinate'))
                rM = Matrix4.get_points_matrix(time_series_dict[closest_down]['value'], frame_coordinate)
                time_series_record[frame_number] = Matrix4.get_points_matrix(rM, down_coordinate)
            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                down_coordinate = self.capture.get_world(self.capture.attachment[int(closest_down)].get('coordinate'))
                up_coordinate = self.capture.get_world(self.capture.attachment[int(closest_up)].get('coordinate'))
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]

                rM_points = Matrix4.get_points_matrix(rM, down_coordinate)
                lM_points = Matrix4.get_points_matrix(lM, up_coordinate)
                time_series_record[frame_number] = {
                    'x': self.capture.inter_polate_number(*_frame, lM_points['x'], rM_points['x']),
                    'y': self.capture.inter_polate_number(*_frame, lM_points['y'], rM_points['y']),
                    'z': self.capture.inter_polate_number(*_frame, lM_points['z'], rM_points['z'])}
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def box2d_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']

            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                Matrix = {
                    'bottomLeft': {
                        'x': self.capture.inter_polate_number(*_frame, lM['bottomLeft']['x'], rM['bottomLeft']['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM['bottomLeft']['y'],
                                                              rM['bottomLeft']['y'])},
                    'bottomRight': {
                        'x': self.capture.inter_polate_number(*_frame, lM['bottomRight']['x'], rM['bottomRight']['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM['bottomRight']['y'], rM['bottomRight']['y'])},
                    'topLeft': {
                        'x': self.capture.inter_polate_number(*_frame, lM['topLeft']['x'], rM['topLeft']['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM['topLeft']['y'], rM['topLeft']['y'])},
                    'topRight': {
                        'x': self.capture.inter_polate_number(*_frame, lM['topRight']['x'], rM['topRight']['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM['topRight']['y'], rM['topRight']['y'])}
                }
                time_series_record[frame_number] = Matrix
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def vertices_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']

            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                vertices = []
                for s in range(len(rM)):
                    vertices.append({
                        'x': self.capture.inter_polate_number(*_frame, lM[s]['x'], rM[s]['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM[s]['y'], rM[s]['y'])})
                time_series_record[frame_number] = vertices
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def points_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                points = []
                for s in range(len(rM)):
                    points.append({
                        'x': self.capture.inter_polate_number(*_frame, lM[s]['x'], rM[s]['x']),
                        'y': self.capture.inter_polate_number(*_frame, lM[s]['y'], rM[s]['y']),
                        'z': self.capture.inter_polate_number(*_frame, lM[s]['z'], rM[s]['z'])})
                time_series_record[frame_number] = points
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def point3d_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                time_series_record[frame_number] = {
                    'x': self.capture.inter_polate_number(*_frame, lM['x'], rM['x']),
                    'y': self.capture.inter_polate_number(*_frame, lM['y'], rM['y']),
                    'z': self.capture.inter_polate_number(*_frame, lM['z'], rM['z'])}
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def line_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']

            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                points = []
                if len(rM) == len(lM):
                    for s in range(len(rM)):
                        points.append({
                            'x': self.capture.inter_polate_number(*_frame, lM[s]['x'], rM[s]['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM[s]['y'], rM[s]['y']),
                        })
                    time_series_record[frame_number] = points
                else:
                    time_series_record[frame_number] = rM
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def point_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']

            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']

                _frame = [closest_down, closest_up, frame_number]
                time_series_record[frame_number] = {
                    'x': self.capture.inter_polate_number(*_frame, lM['x'], rM['x']),
                    'y': self.capture.inter_polate_number(*_frame, lM['y'], rM['y'])}
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])

    def cuboid_time_series(self, data_type, time_series_dict, frame_number):
        time_series_dict = deepcopy(time_series_dict)
        if not time_series_dict:
            return []
        frame_list = list(time_series_dict.keys())
        time_series_record = {}
        closest_down, closest_up = self.find_closest_values(frame_list, frame_number)
        if closest_down:
            if str(frame_number) in frame_list:
                if time_series_dict[str(frame_number)]['type'] == 'appear':
                    time_series_record[frame_number] = time_series_dict[str(frame_number)]['value']
            elif time_series_dict[closest_down]['type'] == 'appear' and (
                    time_series_dict[closest_up]['type'] == 'disappear' if closest_up else True):
                time_series_record[frame_number] = time_series_dict[closest_down]['value']

            elif time_series_dict[closest_down]['type'] == 'appear' and time_series_dict[closest_up][
                'type'] == 'appear':
                rM = time_series_dict[closest_down]['value']
                lM = time_series_dict[closest_up]['value']
                _frame = [closest_down, closest_up, frame_number]
                Matrix = {
                    'front': {
                        'bottomLeft': {
                            'x': self.capture.inter_polate_number(*_frame, lM['front']['bottomLeft']['x'],
                                                                  rM['front']['bottomLeft']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['front']['bottomLeft']['y'],
                                                                  rM['front']['bottomLeft']['y'])},
                        'bottomRight': {
                            'x': self.capture.inter_polate_number(*_frame, lM['front']['bottomRight']['x'],
                                                                  rM['front']['bottomRight']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['front']['bottomRight']['y'],
                                                                  rM['front']['bottomRight']['y'])},
                        'topLeft': {
                            'x': self.capture.inter_polate_number(*_frame, lM['front']['topLeft']['x'],
                                                                  rM['front']['topLeft']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['front']['topLeft']['y'],
                                                                  rM['front']['topLeft']['y'])},
                        'topRight': {
                            'x': self.capture.inter_polate_number(*_frame, lM['front']['topRight']['x'],
                                                                  rM['front']['topRight']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['front']['topRight']['y'],
                                                                  rM['front']['topRight']['y'])}},
                    'back': {
                        'bottomLeft': {
                            'x': self.capture.inter_polate_number(*_frame, lM['back']['bottomLeft']['x'],
                                                                  rM['back']['bottomLeft']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['back']['bottomLeft']['y'],
                                                                  rM['back']['bottomLeft']['y'])},
                        'bottomRight': {
                            'x': self.capture.inter_polate_number(*_frame, lM['back']['bottomRight']['x'],
                                                                  rM['back']['bottomRight']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['back']['bottomRight']['y'],
                                                                  rM['back']['bottomRight']['y'])},
                        'topLeft': {
                            'x': self.capture.inter_polate_number(*_frame, lM['back']['topLeft']['x'],
                                                                  rM['back']['topLeft']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['back']['topLeft']['y'],
                                                                  rM['back']['topLeft']['y'])},
                        'topRight': {
                            'x': self.capture.inter_polate_number(*_frame, lM['back']['topRight']['x'],
                                                                  rM['back']['topRight']['x']),
                            'y': self.capture.inter_polate_number(*_frame, lM['back']['topRight']['y'],
                                                                  rM['back']['topRight']['y'])}},
                }
                time_series_record[frame_number] = Matrix
        setattr(self, data_type, time_series_record[frame_number] if time_series_record.get(frame_number) else [])


class Input(UtilityCategory):
    """input"""

    def __init__(self, data, frame_number):
        super().__init__(data)
        self.key = data['key']
        self.target = frame_number
        if data['input'].get('timeSeries'):
            _ = data['input']
            self.frame_list = np.array(list(_['timeSeries'].keys()), dtype=int).tolist()
            temp = str(self.frame_list[-1])
            if _['timeSeries'][temp]['type'] == 'appear' and self.target >= self.frame_list[-1]:
                self.input = {
                    'type': _['type'],
                    'value': _['timeSeries'][temp]['value']
                }
                if (a := _['timeSeries'][temp].get("teamId", None)):
                    self.input['teamId'] = a
            elif _['timeSeries'][temp]['type'] == 'disappear' and self.target >= self.frame_list[-1]:
                self.input = {
                    'type': _['type'],
                }
            else:
                if self.target in self.frame_list:
                    input_value = _['timeSeries'][str(self.target)].get("value")
                    self.input = {
                        'type': _['type'],
                        'value': input_value if input_value else ""
                    }
                    if (a := _['timeSeries'][temp].get("teamId", None)):
                        self.input['teamId'] = a
                else:
                    self.input = self.time_series(_)
        else:
            self.input = data['input']

    def time_series(self, time_series_dict):
        left, right = FrameCapture.binary_search(self.frame_list, self.target)
        if time_series_dict['timeSeries'][str(left)]['type'] == 'appear':
            return {
                'type': time_series_dict['type'],
                'value': time_series_dict['timeSeries'][str(left)]['value']
            }
        else:
            return {
                'type': time_series_dict['type']}



from enum import Enum
from typing import TypeVar
import numpy as np

_T = TypeVar('_T')


class AnnotationObject:
    def __init__(self, **kwargs) -> None:
        _allow = ["id", "label", "children", "score", "parent", "label_kind", "team_id"]
        for key, val in kwargs.items():
            if val and key in _allow:
                setattr(self, key, val)


class NumberFiled:
    def __init__(self, name) -> None:
        self.name = name

    def __set__(self, instance, value) -> None:
        if not isinstance(value, (float, int, np.int64)):
            raise TypeError("point's coordinate must be a number!")
        instance.__dict__[self.name] = value

    def __get__(self, instance, owner) -> _T:
        return instance.__dict__[self.name]

    def __delete__(self, instance) -> None:
        del instance.__dict__[self.name]


class AnnotationType(Enum):
    Point = 'point'
    Point3D = 'point3d'
    Polygon = 'polygon'
    Quadrilateral = 'quadrilateral'
    Box2D = 'rectangle'
    Box3D = 'box3d'
    Line = 'line'
    Line3D = 'line'
    SemanticSegmentation3D = "semantic_segmentation_3d"
    Video = "video"
    Audio = "audio"
    Input = "input"


if __name__ == '__main__':
    pass



from typing import Dict
import math
import numpy as np
from .helper import decode_cam_sources


def pointcloud2cam(points: np.ndarray, extrin: np.ndarray) -> np.ndarray:
    """
    Project point cloud points to camera coordinate

    Args:
        points:
            np.ndarray with shape (N, 4)
        extrin:
            extrinsics, (4, 4)

    Return:
        points_cam:
            Points in camera coordiante, (4, 4)
    """
    points_cam = np.matmul(points, extrin.T)
    mask = points_cam[:, 2] > 0
    points_cam = points_cam[mask]
    return points_cam


def fisheye_cam2img(pts_on_cam, cam_skew, alpha, cam_intrin):
    """
    Fisheye camera, from camera coordinate to image coordinate

    Args:
        pts_on_cam:

        cam_skew:

        alpha:

        cam_intrin:

    Returns:

    """
    k1, k2, k3, k4 = cam_skew[0], cam_skew[1], cam_skew[2], cam_skew[3]
    xc, yc, zc = pts_on_cam[:, 0], pts_on_cam[:, 1], pts_on_cam[:, 2]
    a, b = xc / (np.abs(zc)), yc / (np.abs(zc))  # N * 1, N * 1
    r = np.array([math.sqrt(a_ ** 2 + b_ ** 2) for a_, b_ in zip(a, b)])  # N * 1
    theta = np.array([math.atan(r_) if zc_ > 0 else math.pi - math.atan(r_) for r_, zc_ in zip(r, zc)])  # N * 1
    theta_d = np.array([theta_ * (1 + k1 * theta_ ** 2 + k2 * theta_ ** 4 + k3 * theta_ ** 6 + k4 * theta_ ** 8) for theta_ in theta])  # N * 1
    scale = np.array([1.0 / r_ if r_ > 1e-8 else 1.0 for r_ in r])
    x_ = np.array([(theta_d_ * scale_) * a_ for theta_d_, scale_, a_ in zip(theta_d, scale, a)]).reshape(-1, 1)
    y_ = np.array([(theta_d_ * scale_) * b_ for theta_d_, scale_, b_ in zip(theta_d, scale, b)]).reshape(-1, 1)
    cam_boxes = np.hstack((x_, y_, np.ones((x_.shape[0], 2))))
    cam_intrin[0][1] = alpha
    img_boxes = np.matmul(cam_boxes, cam_intrin.T)
    return img_boxes[:, :2]


def cam2img(points: np.ndarray, intrin: np.ndarray, type=None, cam_skew=None, alpha=None):
    """
    Project points from camera coordinate to image coordiante

    Args:
        points: (N * 4)
        intrin: (N * 4)
        type: None or str
        cam_skew: (4, )
        alpha: int

    Return:
        pts_in_img: (N, 2)
    """
    if type is None or type == 'Pinhole':
        points_img = np.matmul(points, intrin.T)
        x, y = points_img[0] / points_img[2], points_img[1] / points_img[2]
        pts_in_img = np.stack([x, y], axis=1)
    elif type == 'Fisheye':
        if alpha is None:
            alpha = 0
        pts_in_img = fisheye_cam2img(points, cam_skew, alpha, intrin)
    else:
        raise NotImplementedError
    return pts_in_img


def pointcloud2img(frame_info: Dict, cam_idx: int, points_3d: np.ndarray, img_path=None, res_save_path=None, visual=False):
    """
    Project 3D points on images

    Args:
        frame_info: Dict

        cam_idx: int
            dfsghj
        points_3d: np.ndarray (N, 3)
            ghj

    Return:
        img_pixels: (N, 2)
    """
    cams_info_lst = decode_cam_sources(frame_info)
    pcd_points = np.hstack((points_3d, np.ones((points_3d.shape[0], 1))))  # N * 4
    cam = cams_info_lst[cam_idx]['camera']
    cam2lidar_rt = cam.cam2lidar_rt
    lidar2cam_rt = np.asarray(np.linalg.inv(cam2lidar_rt))
    points_cam = pointcloud2cam(pcd_points, lidar2cam_rt)[:, :3]
    if points_cam.shape[0] == 0:
        return None
    cam_type = cam.camera_type
    img_pixels = []
    if cam_type == 'PinHole':
        for point_cam in points_cam:
            img_pixel = cam._pinhole_3dto2d(point_in_cam=point_cam)
            img_pixels.append(img_pixel)
    elif cam_type == 'Fisheye':
        for point_cam in points_cam:
            img_pixel = cam._fisheye_3dto2d(point_in_cam=point_cam)
            img_pixels.append(img_pixel)
    elif cam_type == 'OmniDirectional':
        for point_cam in points_cam:
            img_pixel = cam._omnidirectional_3dto2d(point_in_cam=point_cam)
            img_pixels.append(img_pixel)

    if visual and img_path is not None:
        import cv2
        img = cv2.imread(img_path)
        h, w = img.shape[0], img.shape[1]
        for (x_, y_) in img_pixels:
            if 0 <= x_ < w and 0 <= y_ < h:
                cv2.circle(img, (int(x_), int(y_)), 1, (0, 0, 255))
        if res_save_path is not None:
            cv2.imwrite(res_save_path, img)
        else:
            print('please identify your visualization result save path')
    return img_pixels



import numpy as np

from .base_annotation import NumberFiled


class Point3D:
    x: NumberFiled('x')
    y: NumberFiled('y')
    z: NumberFiled('z')

    def __init__(self, x, y, z):
        """
        This class defines the concept of 3d geometry point

        Args:
            x: a float number
            y: a float number
            z: a float number

        Returns:
            Point3D instance

        Examples:
            .. code-block:: python

                p = Point(11.12, 345.43, 46,78)
        """
        self.x = x
        self.y = y
        self.z = z

    def to_array(self) -> np.array:
        return np.array([self.x, self.y, self.z])



import os
import re
import math
import numpy as np
from numpy.linalg import norm
from typing import Optional, Union

from stardust.components.ego import Ego


class Vector(np.ndarray):
    def __new__(cls, x=0, y=0, z=0):
        obj = np.array([x, y, z]).view(cls)
        return obj

    def sub(self, other):
        return self - other

    def add(self, other):
        return self + other

    def distance_to_squared(self, other):
        return np.sum((self - other) ** 2)


class Line:
    def __init__(self, start, end):
        self.start = Vector(*start)
        self.end = Vector(*end)

    def clone(self):
        return Line3(self.start, self.end)

    def delta(self):
        return self.end.sub(self.start)


class Plane:
    def __init__(self, normal, constant):
        self.normal = normal  # Surface normal
        self.constant = constant  # Distance from origin to any point on the plane

    def clone(self):
        return Plane(np.copy(self.normal), self.constant)

    def apply_matrix4(self, matrix):
        # Extract the 3x3 normal matrix and compute its inverse transpose
        normal_matrix = np.linalg.inv(matrix[:3, :3]).T
        # Transform and normalize the normal vector
        transformed_normal = normal_matrix @ self.normal
        # Select a point on the plane (along the normal at the stored distance)
        point_on_plane = self.normal * -self.constant
        # Update normal vector
        self.normal = transformed_normal / np.linalg.norm(transformed_normal)
        # Promote the point to homogeneous coordinates
        point_on_plane_homogeneous = np.append(point_on_plane, 1.0)
        # Transform the point
        transformed_point_homogeneous = matrix @ point_on_plane_homogeneous
        # Update the distance from the origin to the plane
        self.constant = -np.dot(self.normal, transformed_point_homogeneous[:3])

        return self

    def distance_to_point(self, point):
        return np.dot(self.normal, point) + self.constant

    def intersect_line(self, line, direction):
        # Compute dot product between the plane normal and line direction
        denominator = np.dot(self.normal, direction)
        if denominator == 0:
            if self.distance_to_point(line.start) == 0:
                return line.start
            return None
        k = -(np.dot(line.start, self.normal) + self.constant) / denominator
        if k < 0 or k > 1:
            return None
        return line.start + direction * k

    def __str__(self):
        return f'normal:{self.normal},constant:{self.constant}'


# Default to True for non-slot structures (inputs have no children)
def empty(slot: dict) -> bool:
    if slot['type'] == 'slotChildren':
        return not slot['slotsChildren']
    elif slot['type'] == 'childrenOnly':
        return not slot['childrenOnly']
    elif slot['type'] == 'slot':
        return not slot['slots']
    else:
        return True


def frame_name(file_name: str, frame_number: Optional[int] = None) -> str:
    timestamp_pattern = r'_\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}[+-]\d{2}:\d{2}'
    file_name = os.path.splitext(file_name)[0]
    if frame_number is not None:
        new_file_name = f'{re.sub(timestamp_pattern, "", file_name)}_{str(frame_number).rjust(4, "0")}.json'
    else:
        new_file_name = f'{re.sub(timestamp_pattern, "", file_name)}.json'
    # return new_file_name.replace('Downloads', 'Desktop')
    return new_file_name


def frame_key(key: str) -> dict:
    if key == 'slotChildren':
        return {'slotsChildren': []}
    elif key == 'childrenOnly':
        return {'childrenOnly': []}
    elif key == 'slot':
        return {'slots': []}


def slot_structure(slot: dict) -> dict:
    type_ = slot['type']
    _ = {
        'key': slot['key'],
        'label': slot['label'],
        'type': type_,
    }
    if slot.get("teamId", None):
        _['teamId'] = slot.get("teamId")

    if not type_ == 'input':
        _.update(frame_key(type_))
    else:
        try:
            _['input'] = {'type': slot['input']['type']}
        except KeyError:
            _['input'] = {'type': slot['inputSpecification']['type']}
    return _


def binary_search(lst: list, target: int) -> tuple:
    left = 0
    right = len(lst) - 1

    while left <= right:
        mid = (left + right) // 2
        if lst[mid] == target:
            return mid
        elif lst[mid] < target:
            left = mid + 1
        else:
            right = mid - 1

    return lst[left - 1], lst[left]


def inter_polate_number(left_index: int, right_index: int, target: int, right, left):
    return left + ((right - left) / (right_index - left_index)) * (target - left_index)


def angle_closer_to(left_rotation: float, right_rotation: float):
    if left_rotation - right_rotation > math.pi:
        return right_rotation + 2 * math.pi
    if left_rotation - right_rotation < -math.pi:
        return right_rotation - 2 * math.pi
    return right_rotation


def get_world(coordinate: dict):
    ego_heading = coordinate["egoHeading"]
    ego_position = coordinate["ego"]
    ego = Ego(
        heading=(ego_heading['x'], ego_heading['y'], ego_heading['z'], ego_heading['w']),
        position=(ego_position['x'], ego_position['y'], ego_position['z'])
    )
    return ego.ego_rt


def point3d_to_world(point3d: list, coordinate: dict, overturn: bool = False):
    assert coordinate, 'coordinate must not be empty'
    world = get_world(coordinate)
    matrix = np.array([[point['x'], point['y'], point['z'], 1] for point in point3d])
    if overturn:
        point3ds_world = np.dot(matrix, world.I)  # Transform to the sensor coordinate system
    else:
        point3ds_world = np.dot(matrix, world)  # Transform to the global coordinate system
    point3ds_world = point3ds_world[:, :3] / point3ds_world[:, 3]
    return [dict(zip('xyz', point)) for point in point3ds_world.tolist()]


def box3d_to_world(box3d: list, coordinate: dict, overturn: bool = False):
    assert coordinate, 'coordinate must not be empty'
    world = get_world(coordinate)
    matrix = np.array(box3d).reshape(4, 4).T
    if overturn:
        point3ds_world = np.dot(matrix, world.I)  # Transform to the sensor coordinate system
    else:
        point3ds_world = np.dot(matrix, world)  # Transform to the global coordinate system
    return np.array(point3ds_world.T).flatten().tolist()


def make_rotation_z(theta: float):
    # Build a 4x4 rotation matrix around the Z axis
    c, s = np.cos(theta), np.sin(theta)
    return np.array([
        [c, -s, 0, 0],
        [s, c, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1]
    ])


def point_angle(point: dict):
    x, _, z = point.values()
    angle = np.arccos(np.abs(x) / np.sqrt(x ** 2 + z ** 2))
    return angle * (z / np.abs(z))


def check_point_in_camera(point: dict, theta_length: float) -> bool:
    angle = point_angle(point)
    min_z, max_z = 0.5, 100
    if -max_z < point['z'] < min_z:
        return False
    if point['z'] > max_z or point['z'] < -max_z:
        return False
    if theta_length * 2 < np.pi:
        return np.pi - theta_length < angle * 2 < np.pi + theta_length
    else:
        return np.pi - angle * 2 <= theta_length


if __name__ == '__main__':
    name = [
        0.08573903915964579,
        -1.385659579033742,
        0,
        0,
        2.0921314438311733,
        0.1294526754651037,
        0,
        0,
        0,
        0,
        1.268364013474737,
        0,
        -3.058281007163033,
        2.165372077096646,
        1.6431105252136202,
        1
    ]
    coord = {
        "ego": {
            "x": 2.1812003147458467,
            "y": -0.19653660649169938,
            "z": -0.10265531758175112
        },
        "egoHeading": {
            "x": -0.0115017067585349,
            "y": 0.0016873531021340058,
            "z": -0.014199167104915787,
            "w": 0.9998316094396476
        }
    }
    res = box3d_to_world(name, coord)
    print(res)



import os
import configparser
from datetime import datetime

now_time = datetime.now().strftime('%Y%m%d%H%M')
absolute_path = os.path.dirname(__file__)
config = configparser.ConfigParser()

config_path = f'{os.path.dirname(os.path.abspath(__file__))}/config.ini'
assert os.path.exists(config_path)
config.read(config_path)


# OSS credentials
class OssUser:
    oss_key = config.get('OssUser', 'key')
    oss_secret = config.get('OssUser', 'secret')
    oss_bucket = config.get('OssUser', 'bucket')
    endpoint = config.get('OssUser', 'endpoint')
    prefix_url = config.get('OssUser', 'prefix_url')


# Rosetta credentials
class RosUser:
    username = int(config.get('RosUser', 'username'))
    password = config.get('RosUser', 'password')


class RosConfig:
    env_prod = config.get('RosConfig', 'env_prod')
    env_dev = config.get('RosConfig', 'env_dev')


class SavePath:
    input_rosetta_path = os.path.join(absolute_path, f'api/static/input')
    export_rosetta_path = os.path.join(absolute_path, f'api/static/star_export')


# MorningStar configuration
class MS:
    domain = "ROSETTA"
    url_auth = "https://portal-uat.rosettalab.top/rosetta-open/token/verify"
    url_export = "https://portal-uat.rosettalab.top/rosetta-open/dataset/export"
    url_create = "https://portal-uat.rosettalab.top/rosetta-open/dataset/slice"
    url_import = "https://portal-uat.rosettalab.top/rosetta-open/dataset/import"



from copy import deepcopy
from matplotlib import pyplot as plt
import numpy as np
import open3d as o3d
from stardust.geometry.ops3d.bbox.structure.bbox_np_ops import points_in_rbbox, rbbox3d_to_bev_corners

"""
    Remove reflector so that the bounding box is closer to the vehicle
"""


def compute_distance(box):
    """
    compute distance from box center to origin
    """
    return np.sqrt(box[0] ** 2 + box[1] ** 2 + box[2] ** 2)


def stardust2kitti(box):
    box[:, -1] = np.pi - box[:, -1]
    return box


def get_rm(heading):
    """
    rotate theta counterclock wise
    """
    cos = np.cos(heading)
    sin = np.sin(heading)
    return np.array([[cos, -sin],
                     [sin, cos]]).astype(np.float64)



# todo
def remove_one_side_up(pts, x1, x2, interval, imagine):

    xprev, xcurr = x1, x1
    pts_num_ratio = []
    count = 0
    while count <= 8:
        count += 1
        xcurr = xprev - interval
        # num of points within xprev and xcurr
        mask = ((pts[:, 0] <= xprev) &
                (pts[:, 0] >= xcurr))
        num = sum(mask)
        if count == 1 and num == 0:
            if imagine:
                x1 -= interval
            break
        if count == 1:
            xprev = xcurr
            prev_num = num
            continue
        ratio = num / prev_num
        pts_num_ratio.append(ratio)
        xprev = xcurr
    for i, r in enumerate(pts_num_ratio):
        if r >= 200:
            x1 -= (i + 1) * interval
            break
    return x1


def remove_one_side_bottom(pts, x1, x2, interval, imagine):
    xprev, xcurr = x2, x2
    pts_num_ratio = []
    count = 0
    while count <= 8:
        count += 1
        xcurr = xprev + interval
        # num of points within xprev and xcurr
        mask = ((pts[:, 0] >= xprev) &
                (pts[:, 0] <= xcurr))
        num = sum(mask)
        if count == 1 and num == 0:
            if imagine:
                x2 += interval
            break
        if count == 1:
            xprev = xcurr
            prev_num = num
            continue
        ratio = num / prev_num
        pts_num_ratio.append(ratio)
        xprev = xcurr
    for i, r in enumerate(pts_num_ratio):
        if r >= 200:
            x2 += (i + 1) * interval
            break
    return x2


# todo
def remove_reflec_bev(pts_bev, x1, x2, interval, imagine):
    """
    Args:
        x1: xmax
        x2: xmin
        interval: the interval of slicing bounding box
        imagine

    """
    x1 = remove_one_side_up(pts_bev, x1, x2, interval, imagine)
    x2 = remove_one_side_bottom(pts_bev, x1, x2, interval, imagine)
    return x1, x2


def exclude_reflector(points: np.ndarray,
                      boxes: np.ndarray,
                      imagine: bool,
                      render_path: str):
    """
    In some training scenario, model will have a higher precision when reflector of vehicles isn't within the bounding box
    This function changes the size of bounding boxes to exclude reflectors
    This function assume there is at least one point in bbox

    Args:
        points:
            point cloud points, (N, 3)
        boxes:
            bounding boxes, (M, 7)
        imagine:
            True means bounding box will imagine there is a reflector manually even though it is invisible from point cloud
            False means bounding box will not imagine there is a reflector manually if it is invisible

    Return:
        New bounding boxes, (M, 7)
    """
    gt_heading = deepcopy(boxes[:, -1])[0]
    kitti_boxes = stardust2kitti(boxes)
    masks = points_in_rbbox(points, kitti_boxes)
    for i in range(kitti_boxes.shape[0]):
        box = kitti_boxes[i]
        box_kittilabel = np.array([box[0], box[1], box[2], box[3], box[4], box[5], box[-1]])
        box_bev_corner = rbbox3d_to_bev_corners(box_kittilabel[np.newaxis, ...])[0].astype(np.float64)
        rm = get_rm(box[-1])
        pts_in_box = points[masks[:, i]]
        # rotate object points so that front of the vehicle heads to the y negative axis
        x, y, z = box[:3]
        pts_in_box -= [x, y, z]
        box_bev_corner -= [x, y]
        # to bev
        pts_in_box_bev = pts_in_box[:, :2]
        # Rotate into vehicle-aligned coordinates
        pts_in_box_bev = pts_in_box_bev @ rm.T
        box_bev_corner = box_bev_corner @ rm.T
        x1, x2 = max(box_bev_corner[:, 0]), min(box_bev_corner[:, 0])
        y1, y2 = max(box_bev_corner[:, 1]), min(box_bev_corner[:, 1])
        newx1, newx2 = remove_reflec_bev(pts_in_box_bev, x1, x2, 0.08, imagine)
        new_car_d = newx1 - newx2
        new_bev_center = np.array([[newx2 + (newx1 - newx2) / 2, 0]])
        # Visualize
        new_bev_corner = np.array([[newx1, y1],
                                   [newx1, y2],
                                   [newx2, y1],
                                   [newx2, y2]])
        render(pts_in_box_bev, box_bev_corner, new_bev_corner, render_path)
        new_bev_center = new_bev_center @ np.linalg.inv(rm).T + [x, y]
        if box[3] < box[4]:
            new_box = [] + new_bev_center.tolist()[0] + [z, new_car_d, box[4], box[5], box[-1]]
        else:
            new_box = [] + new_bev_center.tolist()[0] + [z, box[3], new_car_d, box[5], box[-1]]
        new_box[-1] = gt_heading
        return np.array(new_box)

def order_points(points):
    left_top = points[np.argmin(np.sum(points, axis=1))]
    right_bottom = points[np.argmax(np.sum(points, axis=1))]
    other_points = [point for point in points if (point.tolist() != left_top.tolist() and point.tolist() != right_bottom.tolist())]
    right_top = other_points[0]
    left_bottom = other_points[1]
    ordered_points = np.array([left_top, right_top, right_bottom, left_bottom, left_top])
    return ordered_points

def render(pts_bev, gtbox_bev_corner, procbox_bev_corner, render_path):
    # visualizing
    eval_range = 6
    _, ax = plt.subplots(1, 1, figsize=(10, 10), dpi=500)
    ax.scatter(pts_bev[:, 0], pts_bev[:, 1], c='red', s=0.2)
    gtbox_bev_corner = order_points(gtbox_bev_corner)
    ax.plot(gtbox_bev_corner[:, 0], gtbox_bev_corner[:, 1], linestyle='solid', color='blue')
    procbox_bev_corner = order_points(procbox_bev_corner)
    ax.plot(procbox_bev_corner[:, 0], procbox_bev_corner[:, 1], linestyle='solid', color='green')
    ax.set_xlim(-eval_range - 3, eval_range + 3)
    ax.set_ylim(-eval_range - 3, eval_range + 3)
    plt.axis('off')

    plt.savefig(f"{render_path}/demo.png")
    plt.close()


if __name__ == '__main__':
    gt_boxes = np.array([[-2.113077501102503, -5.521042369327457, 1.0185785336759907, 2.0715079066265627, 4.550000000000131, 1.696969696969697, 2.7185993154604535]])
    pcd_path = 'box_ops/1693197324100.pcd'
    pcd = o3d.io.read_point_cloud(pcd_path)
    pcd_points = np.asarray(pcd.points)
    new_box = exclude_reflector(pcd_points, gt_boxes, imagine=False)



from typing import Optional, List
import numpy as np
import cv2 as cv
from dataclasses import dataclass
from stardust.components.annotations import Box3D, Point3D
from stardust.components.camera.camera import Camera
from stardust.components.attachment.attachment import AttachmentType, UriInfo

ALL_IMAGE = (".jpg", ".jpeg", ".png", ".webp")


class Image(UriInfo):
    def __init__(
            self,
            uri: str,
            width: Optional[int] = None,
            height: Optional[int] = None,
            name: Optional[str] = None,
            camera_param: Optional[Camera] = None
    ):
        """
        Picture object
        
        Args:
            width: 
                the picture width
            height: 
                the picture height
            name: 
                the picture name
            camera_param:
                class Camera
                    
        Returns:
            Image instance
        
        Examples:
            .. code-block:: python
        
                image = Image('', camera_param=cam)
        """
        super(Image, self).__init__(uri)
        self.type = AttachmentType.Image.value
        self.width = width
        self.height = height
        self.name = name
        self.camera_param = camera_param

    @staticmethod
    def _visual(**kwargs):
        img = cv.imread(kwargs['file_path'])
        cv.rectangle(img, kwargs['min_point'], kwargs['max_point'], (0, 255, 0), thickness=3, lineType=cv.LINE_AA)
        cv.imshow('Image', img)
        cv.waitKey(0)

    @staticmethod
    def _visual2(*args):
        box1, box2, file_path = args
        img = cv.imread(file_path)
        for i in range(4):
            cv.line(img, tuple(box1[i]), tuple(box1[(i + 1) % 4]), (0, 255, 0), 3)
            cv.line(img, tuple(box2[i]), tuple(box2[(i + 1) % 4]), (0, 255, 0), 3)
            Image.draw_dashed_line(img, tuple(box1[i]), tuple(box2[i]), (0, 255, 0), 3, line_type=cv.LINE_AA, shift=0)
        cv.imshow('Image', img)
        cv.waitKey(0)
        cv.destroyAllWindows()

    @staticmethod
    def draw_dashed_line(image, start_point, end_point, color, thickness, line_type, shift):
        line_length = int(np.sqrt((end_point[0] - start_point[0]) ** 2 + (end_point[1] - start_point[1]) ** 2))
        x_step = (end_point[0] - start_point[0]) / line_length
        y_step = (end_point[1] - start_point[1]) / line_length

        for i in range(0, line_length, 10):  # Step size of 10; adjust as needed
            x1 = int(start_point[0] + i * x_step)
            y1 = int(start_point[1] + i * y_step)
            x2 = int(start_point[0] + (i + 5) * x_step)
            y2 = int(start_point[1] + (i + 5) * y_step)
            cv.line(image, (x1, y1), (x2, y2), color, thickness, lineType=line_type, shift=shift)

    @staticmethod
    def get_points(max_point, min_point):
        x1, y1 = np.array(min_point, int)
        x2, y2 = np.array(max_point, int)
        rectangle_coordinates = [(x1, y1), (x2, y1), (x2, y2), (x1, y2)]
        return rectangle_coordinates

    def visual_2d_box(self, box3d: Box3D, file_path: Optional[str]):
        vertices = box3d.vertices
        vertices_matrix = np.c_[np.array([vertices[n] for n in [0, 3, 6, 1]]), np.array([1, 1, 1, 1])]
        pixel_points = np.array([self.camera_param.get_pixel_coordinate(_[:3]) for _ in
                                 np.dot(vertices_matrix, self.camera_param.cam2lidar_rt.I.T).tolist()], dtype=np.int32)
        max_point = max(pixel_points, key=lambda x: sum(x))
        min_point = min(pixel_points, key=lambda x: sum(x))
        file_path = file_path if file_path else self.download_url()
        self._visual(file_path=file_path, max_point=max_point, min_point=min_point)

    def visual_3d_box(self, box3d: Box3D, file_path: Optional[str]):
        vertices = box3d.vertices
        vertices_matrix = np.c_[np.array(vertices), np.array([1] * 8)]
        pixel_points = np.array([self.camera_param.get_pixel_coordinate(_[:3]) for _ in
                                 np.dot(vertices_matrix, self.camera_param.cam2lidar_rt.I.T).tolist()])
        box1 = np.array([pixel_points[_] for _ in [0, 1, 3, 6]])
        sum_of_coordinates1 = np.sum(box1, axis=1)
        max_coordinate1 = box1[np.argmax(sum_of_coordinates1)]
        min_coordinate1 = box1[np.argmin(sum_of_coordinates1)]
        box1 = self.get_points(max_coordinate1, min_coordinate1)

        box2 = np.array([pixel_points[_] for _ in [2, 4, 5, 7]])
        sum_of_coordinates2 = np.sum(box2, axis=1)
        max_coordinate2 = box2[np.argmax(sum_of_coordinates2)]
        min_coordinate2 = box2[np.argmin(sum_of_coordinates2)]
        box2 = self.get_points(max_coordinate2, min_coordinate2)

        self._visual2(box1, box2, file_path)

    @staticmethod
    def gen_image(image_obj):
        """
        generate the Image obj

        Args:
            image_obj:
                rosetta image data
        Returns:
            a Image instance

        """
        _ = Image(uri=image_obj['url'],
                  width=image_obj.get('width'),
                  height=image_obj.get('height'),
                  name=image_obj['name'])
        if "camera" in image_obj:
            _.camera_param = Camera.gen_camera(image_obj['camera'])
        return _


@dataclass
class ImageSet:
    """
    Args:
        image_set: Image list
    """
    type = AttachmentType.ImageSet.value
    image_set: List[Image]


if __name__ == '__main__':
    box3d = Box3D(
        center=Point3D(19.617147983589785, -0.3571670183003431, 0.2083393465328241),
        size=Point3D(2.0032044225562577, 4.793909959547556, 1.5418855556278792),
        rotation=[0, 0, -1.5731558280731583]
    )
    cam = Camera(
        'PinHole',
        heading=(0.4351436321314605, -0.4677553765588058, 0.5492505771794342, -0.5386824023335351),
        position=(1.5601294, -0.37986112, 1.3333596),
        intrinsic=(1846.008599353862, 1857.930995475229, 1948.09353722228, 1078.768188621484),
        radial=(0, 0, 0, 0),
        tangential=(0, 0)
    )
    image = Image('', camera_param=cam)
    # https://rosettalab.top/task/970/224049822964001714/viewOnly
    # image.visual_2d_box(box3d, '/Users/stardust/Desktop/1681532060061245.jpg')
    image.visual_3d_box(box3d, '/Users/mac/Downloads/jc93_suz_Ew_0b_sunny_m_0_1701929360184.jpg')



"""
    Get rosetta's data
"""

import os
import time
import requests
from typing import List, Optional

from stardust.rosetta.rosetta_pool import RosettaPool
from stardust.rosetta.frames2frame import frames_split
from stardust.rosetta.new_farmes2frame import to_split
from stardust.file.file import File


class RosettaData:
    def __init__(
            self,
            project_id: int,
            save_path: str,
            rosetta_type: str = 'top',
    ):
        self.project_id = project_id
        assert rosetta_type in ['top', 'dev'], 'top or dev'
        self.rosetta_type = rosetta_type
        self.save_path = save_path
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        self.pool = RosettaPool(self.rosetta_type)

    def _get_rosetta_json(self, project_id: int, pool_id: list):
        """
        Send a request to export data

        Args:
            project_id:
                project_id
            pool_id:
                pool_id
        Returns:
            response.content
        """
        url = f'{self.pool.env}/rosetta-service/project/doneTask/export'

        req_info = {
            "projectId": project_id,
            "poolId": pool_id,
            "type": 1
        }

        resp = requests.post(url, headers=self.pool.headers, json=req_info)
        if (ct := resp.headers.get('Content-Type')) == "application/octet-stream;charset=utf-8":
            return resp.content
        elif ct == "application/json":
            raise ValueError(resp.json()['message'])
        else:
            raise Exception("Failed to download dataset")

    def _zipfile_factory(self, zipfile: bytes, project_id: int, version: str):
        """
        decompression

        Args:
            zipfile:
                response.content
            project_id:
                project_id
            version:
                version
        Returns:

        """
        save_path = os.path.join(self.save_path, f'{project_id}', "json")
        File.to_unzip(zipfile, save_path)
        if version == 'old':
            frames_split(save_path)
        elif version == 'new':
            to_split(save_path)
        else:
            raise ValueError('version is new or old')

    def export(self, pool_list: Optional[List[int]] = None, split_name: str = 'new'):
        """
        Derived data

        Args:
            pool_list:
                Pool list, optional. If this parameter is not specified, all pools are completed
            split_name:
                new or old

        Returns:

        """
        if not pool_list:
            pool_list = self.pool.get_all_finish_pool_info(self.project_id)
            assert pool_list, f'Project {self.project_id} has no completed pools'
        resp_export = self._get_rosetta_json(self.project_id, pool_list)
        self._zipfile_factory(resp_export, self.project_id, split_name)


if __name__ == '__main__':
    data = RosettaData(1692, '/Users/mac/Desktop')
    data.export([42789,], 'new')



from typing import Iterator
from enum import Enum
from uuid import uuid4

from .base_annotation import _T, AnnotationObject


class InputTypeEnum(Enum):
    Select = "select"
    Multiselect = "multiple"
    Nest = "nested"
    Number = "number"
    Text = "text"
    MultiText = "multiple-text"


class Input(AnnotationObject):
    def __init__(self,
                 input_type,
                 value: _T = None,
                 required: bool = False,
                 default: _T = None,
                 name: str = None,
                 options=None,
                 **kwargs
                 ) -> None:
        """
        This structure defines that your input can only be a drop-down box or an input textarea box

        Args:
            kind:
                one type of [select, text, multiple selet]
            value:
                _T
            required:
                if necessary
            default:
                default value
            name:
                the name of the current select obj
            options:
                All choices allowed by value

        Returns:
            Input instance

        Examples:
            .. code-block:: python

                input = Input(input_type="select", value="rain")
                input.value

                output:
                >>> rain
        """
        assert input_type in [member.value for member in InputTypeEnum]
        self.input_type = input_type
        self.name = name
        self.required = required
        if options:
            assert isinstance(options, Iterator)
        self.value = value

        # Required but no default provided
        if required and not default:
            assert value
        # Normalize multiselect/nested values to list form
        if input_type in (InputTypeEnum.Multiselect.value,
                          InputTypeEnum.Nest.value) \
                and not isinstance(self.value, Iterator):
            self.value = [self.value, ]
        # Single-select paths fall through
        else:
            pass
            # assert isinstance(value, (str, int, float, bool))

        AnnotationObject.__init__(self, **kwargs)

    @staticmethod
    def gen_input(child, parent_id=None):
        """
        generate the input obj

        Args:
            child:
                subclass
            parent_id:
                superclass

        Returns:
            a Input instance
        """
        input_id = str(uuid4())
        parent_id = parent_id if parent_id else ""
        input_obj = Input(input_type=child['input']['type'].split("-")[0],
                          value=child['input'].get('value', None),
                          parent=parent_id,
                          name=child['label'],
                          team_id = child['input'].get('teamId', None),
                          id=input_id)
        return input_obj



from enum import Enum
from typing import List, Union, Sequence

import numpy as np

from .point import Point
from .base_annotation import AnnotationObject


class InterpolationEnum(Enum):
    Straight = "line"
    BSpline = "b-spline"
    Spline = "spline"


class Line(AnnotationObject):
    def __init__(self,
                 points: Union[List[Point], np.array, Sequence],
                 interpolation=InterpolationEnum.Straight.value,
                 **kwargs) -> None:
        """
        This class defines the structure of 2d line.

        Args:
            points:
                A list composed of Point objects or a Sequence composed of numbers
            interpolation:
                Filled line between points, default straight line

        Returns:
            Line instance

        Examples:
            .. code-block:: python

                line = Line([(0,0), (1,1), (2,2), (3,3)])
        """
        assert len(points) >= 2
        if isinstance(points[0], Point):
            self.points = [[point.x, point.y] for point in points]
        else:
            assert np.asarray(points).shape[1] == 2
            self.points = points
        if interpolation not in [member.value for member in InterpolationEnum]:
            raise ValueError("your input interpolation is illegal！")
        self.interpolation = interpolation
        AnnotationObject.__init__(self, **kwargs)

    @staticmethod
    def gen_line(slot, children_lst=None, parent_id=None, label_kind=None, **kwargs):
        """
        generate the Box3D obj

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor ID

        Returns:
                a Line instance
        """
        children_lst = children_lst if children_lst else {}
        parent_id = parent_id if parent_id else ""
        label_kind = label_kind if label_kind else ""
        points = np.array([[point['x'], point['y']] for point in slot['vertices']])
        line = Line(
            id=slot['id'],
            points=points,
            parent=parent_id,
            label=label_kind,
            children=children_lst,
            team_id=kwargs.get("team_id", None)
        )
        return line



from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd

# from stardust.utils.convert import read_rosetta


def get_demo():
    """

    Returns: pandaset format

    """
    columns = ['uuid', 'label', 'yaw', 'stationary', 'camera_used', 'position.x',
               'position.y', 'position.z', 'dimensions.x', 'dimensions.y',
               'dimensions.z', 'attributes.object_motion', 'cuboids.sibling_id',
               'cuboids.sensor_id', 'attributes.pedestrian_behavior',
               'attributes.pedestrian_age', 'attributes.rider_status']
    demo = {key: np.NaN for key in columns}
    demo["camera_used"] = -1
    demo["cuboids.sensor_id"] = -1
    demo["cuboids.sibling_id"] = '-'
    return demo


def _frame_handle(frame_obj, save_path: Path):
    """

    Args:
        frame_obj: sdk frame object
        save_path: data saving path

    Returns:

    """

    task_id = frame_obj.task_info.task_id
    frame_num = frame_obj.task_info.frame_num
    res = []
    input_objs = frame_obj.annotation.input_lst
    for box_id, box_3d in frame_obj.annotation.box3d_lst.items():
        demo = get_demo()
        demo["uuid"] = box_id
        demo["yaw"] = box_3d.rotation[-1]
        demo['position.x'], demo['position.y'], demo['position.z'] = box_3d.center
        demo['dimensions.x'], demo['dimensions.y'], demo['dimensions.z'] = box_3d.size
        for child_id in box_3d.children:
            if input_objs.get(child_id) is None:
                continue
            children_obj = input_objs[child_id]
            attribute_name = children_obj.name
            if attribute_name not in demo.keys():
                continue
            value = children_obj.value
            if value == "False":
                value = False
            elif value == "True":
                value = True
            demo[attribute_name] = value
        res.append(demo)
    output_dir = save_path / f"pandaset/{task_id}/annotations/cuboids"
    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    df = pd.DataFrame(res)
    df.to_pickle(output_dir / f"{frame_num:0>2}.pkl")


def to_pandaset(sdk_generator, save_path: Union[Path, str]):
    """
    Args:
        sdk_generator: sdk builder
        save_path: save path

    Returns: None

    """

    if not isinstance(save_path, Path):
        save_path = Path(save_path)
    for frame_obj in sdk_generator:
        _frame_handle(frame_obj, save_path)


if __name__ == '__main__':
    rosetta_data = '/Users/mac/Downloads/rosetta_data'
    project_id = 2661
    # export rosetta data and cut frames
    sd_generator = read_rosetta(project_id=project_id,
                             input_path=rosetta_data,
                             )
    to_pandaset(sd_generator, Path('/Users/mac/Downloads/rosetta_data'))



import os
import shutil
from io import BytesIO
from zipfile import ZipFile


class File:
    def __init__(self):
        pass

    @staticmethod
    def mkdir(path: str):
        if os.path.isfile(path):
            dir_path = os.path.dirname(path)
            os.makedirs(dir_path, exist_ok=True)
        else:
            dir_path = path
            os.makedirs(path, exist_ok=True)
        return dir_path

    @staticmethod
    def rmfile(path: str):
        os.remove(path)

    @staticmethod
    def rmdir(path: str):
        shutil.rmtree(path)

    @staticmethod
    def to_unzip(byte_data: bytes, path: str):
        zip_data = BytesIO(byte_data)
        z_file = ZipFile(zip_data, 'r')
        for zf in z_file.namelist():
            z_file.extract(zf, path)
        z_file.close()
        return path


if __name__ == '__main__':
    path = '/Users/stardust/Desktop/shangqi'
    File.rmdir(path)



import os
from dataclasses import dataclass
from typing import List, Optional

import numpy as np
import open3d as o3d

from stardust.components.attachment.attachment import AttachmentType, UriInfo

ALL_POINTCLOUD = (".pcd", ".ply")


class PointCloud(UriInfo):
    def __init__(
            self,
            uri: Optional[str] = None,
            name: Optional[str] = None,
            file_path: Optional[str] = None
    ):
        """
        Point cloud structure, marked as point cloud type

        Args:
            uri:
                Cloud address of the point cloud
            name：
                Point cloud name
            file_path:
                Local path of the point cloud file

        Returns:
            PointCloud instance
        
        Examples:
            .. code-block:: python
        
                path = '/Users/stardust/Desktop/1681532060081818.pcd'
                p = PointCloud(filename=path)
                print(p.points_info)
        """

        super(PointCloud, self).__init__(uri)
        self.type = AttachmentType.PointCloud.value
        self.name = name
        # assert os.path.isfile(file_path)
        _, ext = os.path.splitext(file_path)
        assert ext
        self.file_path = file_path
        _ = os.path.splitext(os.path.basename(self.file_path))[0]
        # assert _.isdigit(), 'Name the point cloud file according to the time stamp'
        self.timestamp = _

    @property
    def pcd(self):
        return o3d.t.io.read_point_cloud(self.file_path)

    def points_info(self, add1=False) -> np.ndarray:
        _ = self.pcd.point.positions.numpy()
        if add1:
            _ = np.c_[_, np.array([1] * len(_))]
        return _

    def get_attribute_items(self, attribute: str) -> Optional[np.ndarray]:
        try:
            return eval(f'self.pcd.point.{attribute}.numpy()')
        except KeyError as e:
            print(f'No {attribute} attribute, point cloud information:')
            print(self.pcd)
            return None


@dataclass
class PointCloudSet:
    """
    Args:
        point_cloud_set:
            PointCloud list

    Returns:
        PointCloudSet class
    """
    type = AttachmentType.PointCloudSet.value
    point_cloud_set: List[PointCloud]


if __name__ == '__main__':
    path = '/Users/stardust/Desktop/files/000206.pcd'
    p = PointCloud(file_path=path)
    print(p.points_info(True))
    print(p.__dir__())



from typing import List, Union, Sequence
import numpy as np
from shapely.geometry import Polygon as SPolygon
from .point import Point
from .base_annotation import AnnotationType, AnnotationObject
from .line import Line
from .box import Box2D


class Polygon(Line, AnnotationObject, SPolygon):
    def __init__(self,
                 points: Union[List[Point], np.ndarray, Sequence],
                 **kwargs) -> None:
        """
        This class defines the concept of 2d geometry Polygon

        Args:
            points:
                A list composed of Point objects or a Sequence composed of numbers

        Returns:
            Polygon instance

        Examples:
            .. code-block:: python

                polygon = Polygon([(0,0), (1,1), (2,2), (3,3)])
        """
        if not len(points) > 2:
            raise ValueError("Invalid number length of points")
        if isinstance(points[0], Point):
            self.shapely_polygon = SPolygon([[point.x, point.y] for point in points])
        else:
            assert np.array(points).shape[1] == 2
            self.shapely_polygon = SPolygon([[point[0], point[1]] for point in points])

        Line.__init__(self, points=points)
        AnnotationObject.__init__(self, **kwargs)

    @property
    def area(self) -> float:
        """
        Returns: float
            the area of current polygon instance
        """
        return self.shapely_polygon.area

    @property
    def points_num(self) -> float:
        """
        Returns: float
             the points num of current polygon instance
        """
        return len(self.points)

    @property
    def bounding_box(self):
        """
        Returns: float
            box2d obj, the box3d outer the current polygon
        """
        bounds = self.shapely_polygon.bounds
        return Box2D(np.array([bounds[0], bounds[1]]), np.array([bounds[2], bounds[3]]))

    @staticmethod
    def gen_polygon(slot, children_lst, parent_id, label_kind):
        """
        generate the Polygon obj

        Args:
            slot:
                rosetta slot
            children_lst:
                rosetta children
            parent_id:
                Upper floor ID

        Returns:
            a Polygon instance
        """
        children_lst = children_lst if children_lst else {}
        parent_id = parent_id if parent_id else ""
        label_kind = label_kind if label_kind else ""
        polygon = Polygon(
            points=np.asarray([[round(vertice['x'], 2), round(vertice['y'], 2)] for vertice in slot['vertices']]),
            parent=parent_id,
            id=slot['id'],
            label=label_kind,
            children=children_lst
        )
        return polygon



import math
import json
import os
from typing import List, Union, Generator

import cv2
import numpy as np
from scipy.optimize import linear_sum_assignment

from stardust.components.annotations.point import Point
from stardust.components.annotations.point3d import Point3D


class LaneIoU2D:
    def __repr__(self):
        description = 'Calculate IOU of lanes'
        return description

    def _drawlane(self, lane: List, img_size: tuple, line_width: int):
        img = np.zeros(img_size, dtype=np.uint8)

        for p1, p2 in zip(lane[:-1], lane[1:]):
            cv2.line(img,
                     (int(p1[0]), int(p1[1])),
                     (int(p2[0]), int(p2[1])),
                     color=(255, 255, 255),
                     thickness=line_width)
        return img

    def _intersection_over_union_lane_2d(self, gt_lanes: List[Point], pd_lanes: List[Point], img_size: tuple):
        """
        Compute IntersectionOverUnion of two 2d lanes

        Args:
            gt_lanes: List[Point]
                The ground true of lanes
            pd_lanes: List[Point]
                The prediction of lanes
            img_size: tuple
                height and width of origin data

        Returns:
            np.ndarray: 
                IoU of 2D lanes

        Examples:
            .. code-block:: python

                from stardust.metric.lane_detection import LaneIoU2D
                metric = LaneIoU2D()
                gt_lanes = [Point(1, 1), Point(2, 2), Point(3, 3)]
                pd_lanes = [Point(1, 1), Point(2, 2), Point(3, 3)]
                img_size = [256, 512]
                IOU = metric._intersection_over_union_lane_2d(gt_lanes, pd_lanes, img_size)
        """
        gtlane = [self._drawlane(lane.points.tolist(), img_size=img_size, line_width=10) > 0 for lane in list(gt_lanes.values())]
        predlane = [self._drawlane(lane.points.tolist(), img_size=img_size, line_width=19) > 0 for lane in list(pd_lanes.values())]
        IoUs = np.zeros((len(gtlane), len(predlane)))
        for i, x in enumerate(gtlane):
            for j, y in enumerate(predlane):
                IoUs[i, j] = (x & y).sum() / ((x | y).sum() + 1e-6)
        return IoUs

    def _keypoints_metric(self, gt_points: List[Point], pred_points: List[Point], dist_threshold: float):
        """
        Compute KeypointsMetric of two 2d lanes

        Args:
            gt_lanes: List[Point]
                The ground truth of points
            predlane: List[Point]
                The prediction of points
            dist_threshold: float
                The dist threshold of tp points

        Returns:
            int: 
                keypoint metric of lanes

        Examples:
            .. code-block:: python

                from stardust.metric.lane_detection import LaneIoU2D
                metric = LaneIoU2D()
                gt_lanes = [Point(1, 1), Point(2, 2), Point(3, 3)]
                pd_lanes = [Point(1, 1), Point(2, 2), Point(3, 3)]
                tp = metric.KeypointsMetric(gt_lanes, pd_lanes, 3)
        """
        point_distance = np.zeros((len(pred_points), len(gt_points)))
        # TODO: replace with vectorized matrix operations
        for row, pred_point in enumerate(pred_points):
            pred_pt = np.array([pred_point[0], pred_point[1]])
            for col, gt_point in enumerate(gt_points):
                gt_pt = np.array([gt_point[0], gt_point[1]])
                point_distance[row][col] = np.linalg.norm(gt_pt - pred_pt)
        dist_list = []
        for row in range(point_distance.shape[0]):
            ind = np.argpartition(point_distance[row], -2)[:2]
            nearest_points0, nearest_points1 = gt_points[ind[0]], gt_points[ind[1]]
            gt_p0_x, gt_p0_y, gt_p1_x, gt_p1_y = nearest_points0[0], nearest_points0[1], nearest_points1[0], nearest_points1[1]
            if gt_p1_x != gt_p0_x:
                k = (gt_p1_y - gt_p0_y) / (gt_p1_x - gt_p0_x)
                b = gt_p1_y - k * gt_p1_x
                A, B, C = k, -1, b
            else:
                A, B, C = 1, 0, -gt_p1_x
            pred_p_x, pred_p_y = pred_points[row][0], pred_points[row][1]
            dist = abs(A * pred_p_x + B * pred_p_y + C) / math.sqrt(A ** 2 + B ** 2)
            dist_list.append(dist)
        tp = int(np.sum(np.where(np.array(dist_list) < dist_threshold, True, False)))
        return tp

    # def IntersectionOverUnion3DLane(self, gtlane: List[Point3D], predlane: List[Point3D]):
    #     pass

    # todo
    def compute(self, gt_lanes: Union[List[Point], List[Point3D]], pd_lanes: Union[List[Point], List[Point3D]], IoU_thr: float, kpt_thr: float, img_size: tuple, lane_type: str):
        """
        Compute metric of lanes

        Args:
            gt_lanes: List[Point]
                The ground truth of lanes
            pd_lane: List[Point]
                The prediction of lanes
            IoU_thr: float
                The IoU threshold of tp lanes
            kpt_thr: float
                The dist threshold of tp points
            img_size: tuple
                height and width of origin data
            lane_type: str
                which type of lane to compute metric, only support '2D' for now

        Returns:
            Tuple: 
                metric of lane_gt, lane_pd, lane_tp, kpt_pd, kpt_tp

        Examples:
            .. code-block:: python

                from stardust.metric.lane_detection import LaneIoU2D
                metric = LaneIoU2D()
                gt_lanes = [Point(1, 1), Point(2, 2), Point(3, 3)]
                pd_lanes = [Point(1, 1), Point(2, 2), Point(3, 3)]
                lane_gt, lane_pd, lane_tp, kpt_pd, kpt_tp = laneIOU.compute(gt_lanes, pd_lanes, 0.5, (256, 512), '2D)
        """
        if lane_type == '2D':
            img_size = (1280, 800)
            lane_gt = len(gt_lanes)
            lane_pd = len(pd_lanes)
            kpt_pd = 0
            IOU = self._intersection_over_union_lane_2d(gt_lanes=gt_lanes, pd_lanes=pd_lanes, img_size=img_size)
            row_ind, col_ind = linear_sum_assignment(1 - IOU)
            kpt_tp = 0
            lane_tp = int((IOU[row_ind, col_ind] > IoU_thr).sum())
            for row, col in zip(row_ind, col_ind):
                if IOU[row, col] > IoU_thr:
                    gt_points = list(gt_lanes.values())[row].points.tolist()
                    pd_points = list(pd_lanes.values())[col].points.tolist()
                    kpt_pd += len(pd_points)
                    pt_tp = self._keypoints_metric(gt_points, pd_points, kpt_thr)
                    kpt_tp += pt_tp

            return lane_gt, lane_pd, lane_tp, kpt_pd, kpt_tp
        # else:
        #     IOU = self.IntersectionOverUnion3DLane(gt_lanes, pd_lanes)
        # return metric


IoUMode = {
    '2D': LaneIoU2D()
}


def compute_metric_single_frame(gt_lanes: List, pd_lanes: List, IoU_thr: float, kpt_thr: float, img_size: tuple, lane_type: str):
    """
    Compute metric of all lanes of single frame

    Args:
        gt_lanes: List
           The ground truth of lanes 
        pr_lane: List
           The prediction of lanes 
        IoU_thr: float
            The IoU threshold of tp lanes
        kpt_thr: float
            The dist threshold of tp points
        img_size: tuple
            height and width of origin data
        lane_type: str
            which type of lane to compute metric, only support '2D' for now

    Returns:
        metric: tuple
    """
    if lane_type == '2D':
        assert img_size

    laneIOU = IoUMode[lane_type]
    lane_gt, lane_pd, lane_tp, kpt_pd, kpt_tp = laneIOU.compute(gt_lanes, pd_lanes, IoU_thr, kpt_thr, img_size, lane_type)
    return lane_gt, lane_pd, lane_tp, kpt_pd, kpt_tp


def compute_metric(data: Generator, IoU_thr=0.5, kpt_thr=3, save_path=None):
    """
    Compute metric of lanes

    Args:
        data: Generator
            A generator object to get all information from all frames
        IoU_thr: float
            The IoU threshold of tp lanes
        kpt_thr: float
            The dist threshold of tp points
        save_path: str
            Local path to save metric results

    Returns:Tuple
        The first one represents the metric of every single frame and the second represents the metric of all frames

    Examples:
        .. code-block:: python

            from stardust.metric.lane_detection import compute_metric
            from stardust.rosetta.rosetta_data import RosettaData

            project_id = 856
            json_datas = read_rosetta(project_id=project_id,
                                    input_path=input_path,
                                    )
            metric = compute_metric(json_datas, 0.5, 3, 'local/')
    """
    total_lane_gt_2d = total_lane_pd_2d = total_lane_tp_2d = total_kpt_pd_2d = total_kpt_tp_2d = 0
    metric_output = {}
    for task_id, json_data in enumerate(data):
        gts_2d = json_data.annotation.line_lst
        pds_2d = json_data.prediction.line_lst
        img_size = (json_data.media.image.height, json_data.media.image.width)
        lane_gt, lane_pd, lane_tp, kpt_pd, kpt_tp = compute_metric_single_frame(gts_2d, pds_2d, IoU_thr, kpt_thr, img_size, '2D')
        lane_recall = float('nan') if lane_gt == 0 else lane_tp / lane_gt
        lane_precision = float('nan') if lane_pd == 0 else lane_tp / lane_pd
        lane_f1 = float('nan') if lane_recall + lane_precision == 0 else 2 * lane_recall * lane_precision / (lane_recall + lane_precision)
        kpt_accuracy = float('nan') if kpt_pd == 0 else kpt_tp / kpt_pd
        metric_output[task_id] = dict(lane_gt=lane_gt, lane_pd=lane_pd, lane_tp=lane_tp, lane_recall=lane_recall, lane_precision=lane_precision, lane_f1=lane_f1, kpt_pd=kpt_pd, kpt_tp=kpt_tp,
                                      kpt_accuracy=kpt_accuracy)
        total_lane_gt_2d += lane_gt
        total_lane_pd_2d += lane_pd
        total_lane_tp_2d += lane_tp
        total_kpt_pd_2d += kpt_pd
        total_kpt_tp_2d += kpt_tp

    total_lane_recall_2d = float('nan') if total_lane_gt_2d == 0 else total_lane_tp_2d / total_lane_gt_2d
    total_lane_precision_2d = float('nan') if total_lane_pd_2d == 0 else total_lane_tp_2d / total_lane_pd_2d
    total_lane_f1_2d = float('nan') if total_lane_recall_2d + total_lane_precision_2d == 0 else 2 * total_lane_recall_2d * total_lane_precision_2d / (total_lane_recall_2d + total_lane_precision_2d)

    total_kpt_accuracy_2d = float('nan') if total_kpt_pd_2d == 0 else total_kpt_tp_2d / total_kpt_pd_2d

    metric_total = dict(total_lane_gt_2d=total_lane_gt_2d, total_lane_pd_2d=total_lane_pd_2d, total_lane_tp_2d=total_lane_tp_2d, total_lane_recall_2d=total_lane_recall_2d,
                        total_lane_precision_2d=total_lane_precision_2d, total_lane_f1_2d=total_lane_f1_2d, total_kpt_pd_2d=total_kpt_pd_2d, total_kpt_tp_2d=total_kpt_tp_2d,
                        total_kpt_accuracy_2d=total_kpt_accuracy_2d)

    if save_path is not None:
        os.makedirs(os.path.join(save_path, 'metric', 'lane_detection'), exist_ok=True)
        with open(os.path.join(save_path, 'metric', 'lane_detection', 'metric_by_task_id.json'), 'w') as f:
            json.dump(metric_output, f)
        with open(os.path.join(save_path, 'metric', 'lane_detection', 'metric_summary.json'), 'w') as f:
            json.dump(metric_total, f)
    # return metric_output, metric_total



import time
import hashlib
from typing import Dict, List
from functools import wraps
import pickle
import requests

from stardust.config import MS as ms_config
from stardust.convertion.to_pandaset import to_pandaset
from stardust.rosetta.frames2frame import func_ms


def decoration(func_obj):
    """
    This decorator is mainly used to invalidate the method auth due to time-outs
    Returns:
    """

    @wraps(func_obj)
    def _func(ms, *args, **kwargs):
        try:
            ret = func_obj(ms, *args, **kwargs)
        except requests.exceptions.ConnectionError as e:
            raise requests.exceptions.ConnectionError("MorningStar connection failed")
        if ret.status_code == 502:
            setattr(ms, "auth", ms._auth())
            ret = func_obj(*args, **kwargs)
        return ret

    return _func


class MS:
    def __init__(self):
        self.config = ms_config
        self.app_key = f"{ms_config.domain}_OPEN_APP_KEY"
        self.auth = self._auth()

    def _md5(self, input_string) -> str:
        """
        Generates md5 digest of MS, dependent on time and OPEN_APP_KEY
        Args:
            input_string:

        Returns:
            str
        """
        return hashlib.md5(input_string.encode()).hexdigest()

    def _auth(self):
        self._time = time_value = int(time.time() * 1000)
        self.md5_value = self._md5(f"{str(time_value)}{self.app_key}")
        payload = {
            "appKey": self.app_key,
            "sign": self.md5_value,
            "createTime": self._time
        }
        resp = requests.post(self.config.url_auth,
                             headers={
                                 "Content-Type": "application/json"},
                             json=payload)
        assert resp.json()['code'] == 20, "Authentication failed"

    def _export_dataset_ms(self, dataset_id: int,
                           version_num: int,
                           slice_id: int = None,
                           page_no: int = 1,
                           page_size: int = 10
                           ) -> requests.Response:
        assert dataset_id and version_num
        url = self.config.url_export
        req_json = {
            "datasetId": dataset_id,
            "versionNum": version_num,
            "sliceId": slice_id,
            "pageNo": page_no,
            "pageSize": page_size,
            "appKey": self.app_key,
            "sign": self.md5_value,
            "createTime": self._time,
        }
        res = requests.post(url,
                            headers={
                                "Content-Type": "application/json"
                            },
                            json=req_json)
        return res

    def _create_dataset_ms(self, dataset_id: int,
                           data_instance_ids: List[str],
                           name: str,
                           description: str,
                           version_num: int
                           ) -> requests.Response:
        url = self.config.url_create
        req_json = {
            "datasetId": dataset_id,
            "dataInstanceIds": data_instance_ids,
            "appKey": self.app_key,
            "sign": self.md5_value,
            "createTime": self._time,
            "name": name,
            "description": description,
            "versionNum": version_num,
        }
        res = requests.post(url,
                            headers={
                                "Content-Type": "application/json"
                            },
                            json=req_json)
        return res

    def _import_dataeet_slice(self,
                              dataset_id: int,
                              model_id: int,
                              version_num: int,
                              annotation_result: List
                              ) -> requests.Response:
        url = self.config.url_import
        req_json = {
            "datasetId": dataset_id,
            "modelId": model_id,
            "versionNum": version_num,
            "appKey": self.app_key,
            "sign": self.md5_value,
            "createTime": self._time,
            "annotationResult": annotation_result
        }
        res = requests.post(url,
                            headers={
                                "Content-Type": "application/json"
                            },
                            json=req_json)
        return res

    # Export dataset content
    def export_dataset(self, **kwargs) -> Dict:
        """
        Export ms data and automatically detach consecutive frames

        Args:
            dataset_id: int
                Data set ID
            version_num: int
                Version number
            slice_id: int
                Slice ID
            page_no: int
                Slice paging
            page_size: int
                Amount of data per page

        Returns: Dict
            Derived data slicing

        Examples:
            .. code-block:: python

                from stardust.ms.ms import MS

                frame_gen = MS().export_dataset(
                    dataset_id=351787480925605888,
                    version_num=18
                )
                for frame in frame_gen:
                    pass

        """
        # Request dataset export from MorningStar
        resp = self._export_dataset_ms(**kwargs)
        assert isinstance(resp, requests.Response) and resp.status_code == 200, "Request failed"
        assert resp.json()['code'] == 2000, f"Dataset export failed, {resp.json()['message']}"

        # Split frames as needed
        ms_data = func_ms(resp.json())

        # Return generator to converter
        return ms_data

    # Create dataset slice
    def create_dataset(self, **kwargs) -> Dict:
        """
        Create a dataset slice

        Args:
            dataset_id: int
                Data set ID
            data_instance_ids: int
                Data instance ID
            name: int
                Slice name
            description: int
                Slice description

        Returns: Dict
            Create slice results

        Examples:
            .. code-block:: python

                from stardust.ms.ms import MS

                resp = MS().create_dataset(
                    dataset_id=352036425840988160,
                    data_instance_ids=[351787490434093056],
                    name="Slice 1",
                    description="description"
                )
                print(resp)

        """
        resp = self._create_dataset_ms(**kwargs)
        assert isinstance(resp, requests.Response) and resp.status_code == 200, "Request failed"
        assert resp.json()['code'] == 2000, f"Dataset creation failed, {resp.json()['message']}"
        return resp.json()

    # Import pre-processed slice
    def import_dataset_ms(self, **kwargs) -> Dict:
        """
        Import data set

        Args:
            dataset_id: int
                Data set ID
            model_id: int
                Model ID
            version_num: int
                Version number
            annotation_result: List
                Result of pretreatment

        Returns: Dict
            Import result

        Examples:
            .. code-block:: python

                from stardust.ms.ms import MS

                resp = MS().import_dataset_ms(
                    dataset_id=351787480925605888,
                    model_id="404",
                    version_num=1,
                    annotation_result=[
                          {
                            "annotation": {
                              "annotations": [
                                {
                                  "key": "3D box",
                                  "label": "3D box",
                                  "type": "slotChildren",
                                  "slotsChildren": [...]
                                }
                              ],
                              "operators": [...]
                            },
                            "dataInstanceId": "351787490434093056"
                          },
                          {
                            "annotation": {
                              "annotations": [...],
                              "operators": [...]
                            },
                            "dataInstanceId": "351787490622836736"
                          }
                        ]
                )
                print(resp)

        """
        resp = self._import_dataeet_slice(**kwargs)
        assert isinstance(resp, requests.Response) and resp.status_code == 200, "Request failed"
        assert resp.json()['code'] == 2000, f"Dataset import failed, {resp.json()['message']}"
        return resp.json()


if __name__ == "__main__":
    # Example export from MorningStar
    gen_data = MS().export_dataset(
        dataset_id=354992337446768640,
        version_num=3
    )
    for data in gen_data:
        print(data)



import collections
import os
import json
from typing import Dict, List, Generator

import numpy as np
from scipy.optimize import linear_sum_assignment

from stardust.components.annotations.box3d import Box3D
from stardust.components.annotations.point3d import Point3D
from stardust.components.annotations.box import Box2D
from stardust.components.annotations.point import Point
from stardust.metric.object_detection import BoxIou2D, BoxIou3D


class TrackingData:
    def __init__(self, track_id: int, box_type: str, obj_info: List):
        """
        Define a tracking object

        Args:
            track_id: int
                tracking id of object
            box_type: str
                type of object, which should be chosen from '2D' and '3D'
            obj_info: List
                info of box, [x, y, z, l, w, h, ry] for box 3d and [x, y, w, h] for box 2d

        Returns:
            TrackingData object

        Examples:
            .. code-block:: python

                from stardust.metric.tracking import TrackingData
                t_data = TrackingData(0, '3D', [1, 2, 3, 4.5, 1.8, 1.5, 0.1])
                t_data = TrackingData(0, '2D', [1, 2, 4.5, 1.8])

        """
        if box_type is None or box_type not in ['2D', '3D']:
            raise ValueError(f"Got invalid box_type {box_type}")
        self.track_id = track_id
        if box_type == '2D':
            if len(obj_info) != 4:
                raise ValueError(f"The input {box_type} box should be [x, y, w, h]")

            self.x = obj_info[0]
            self.y = obj_info[1]
            self.w = obj_info[2]
            self.h = obj_info[3]
        elif box_type == '3D':
            if len(obj_info) != 7:
                raise ValueError(f"The input {box_type} box should be [x, y, z, l, w, h, ry]")

            self.x = obj_info[0]
            self.y = obj_info[1]
            self.z = obj_info[2]
            self.l = obj_info[3]
            self.w = obj_info[4]
            self.h = obj_info[5]
            self.yaw = obj_info[6]
        self.valid = False
        self.tracker = None


class TrackingMetric:
    def __repr__(self):
        description = 'Calculate tracking metric'
        return description

    def __init__(self, trajectory_gt: List, trajectory_pd: List, IoU_thr: float, metric_tyoe: str) -> None:
        """
        Define a TrackMetric object and prepare metric data

        Args:
            trajectory_gt: List
                trajectory of annotation
            trajectory_pd: List
                trajectory of prediction
            IoU_thr: float
                The iou threshold of tp boxes
            metric_tyoe: str
                type of input object, which should be chosen from '2D' and '3D'

        Returns:
            TrackingMetric obj

        """
        self.trajectory_gt = trajectory_gt
        self.trajectory_pd = trajectory_pd
        self.IoU_thr = IoU_thr
        self.metric_tyoe = metric_tyoe
        self.gt = 0
        self.pd = 0
        for frame_id in trajectory_gt:
            self.gt += len(trajectory_gt[frame_id])
        for frame_id in trajectory_pd:
            self.pd += len(trajectory_pd[frame_id])
        self.tp = 0
        self.id_switches = 0
        self.MOTA = 0
        self.ious = []

    def compute(self):
        """
        compute tracking metric

        Args:

        Returns:
            Dict
        """
        max_cost = 1e9
        self.seq_trajectories = collections.defaultdict(list)

        for f in range(len(self.trajectory_gt)):
            g = self.trajectory_gt[f]
            t = self.trajectory_pd[f]
            cost_matrix_iou = []
            iou = np.zeros((len(g), len(t)))
            for gi, gg in enumerate(g):
                gg.tracker = -1
                cost_row_iou = []
                for ti, tt in enumerate(t):
                    if self.metric_tyoe == '3D':
                        pred_bbox = Box3D(Point3D(tt.x, tt.y, tt.z), Point3D(tt.l, tt.w, tt.h), [0, 0, tt.yaw])
                        gt_bbox = Box3D(Point3D(gg.x, gg.y, gg.z), Point3D(gg.l, gg.w, gg.h), [0, 0, gg.yaw])
                        iou[gi][ti] = BoxIou3D().compute_IoU(gt_bbox, pred_bbox, 'IoU')
                    else:
                        pred_bbox = Box2D(center=Point(tt.x, tt.y), size=[tt.w, tt.h])
                        gt_bbox = Box2D(center=Point(gg.x, gg.y), size=[gg.w, gg.h])
                        iou[gi][ti] = BoxIou2D().compute_IoU(gt_bbox, pred_bbox, 'IoU')
                    if iou[gi][ti] >= self.IoU_thr:
                        cost_row_iou.append(1 - iou[gi][ti])
                    else:
                        cost_row_iou.append(max_cost)

                self.seq_trajectories[gg.track_id].append(-1)
                cost_matrix_iou.append(cost_row_iou)

            if len(g) == 0:
                cost_matrix_iou = [[]]
            row_inds, col_inds = linear_sum_assignment(cost_matrix_iou)
            for row, col in zip(row_inds, col_inds):
                c_iou = float(cost_matrix_iou[row][col])
                if c_iou < max_cost:
                    self.ious.append(1 - c_iou)
                    g[row].tracker = t[col].track_id
                    t[col].valid = True
                    self.tp += 1
                    self.seq_trajectories[g[row].track_id][-1] = t[col].track_id
                else:
                    g[row].tracker = -1
        self.fn = self.gt - self.tp
        self.fp = self.pd - self.tp
        for g in self.seq_trajectories.values():
            last_id = g[0]
            for f in range(1, len(g)):
                if last_id != g[f] and last_id != -1 and g[f] != -1 and g[f - 1] != -1:
                    self.id_switches += 1

        if self.gt == 0:
            self.MOTA = float('-inf')
        else:
            self.MOTA = 1 - (self.fn + self.fp + self.id_switches) / float(self.gt)
        self.recall = float('nan') if self.gt == 0 else self.tp / self.gt
        self.precision = float('nan') if self.pd == 0 else self.tp / self.pd
        self.f1 = float('nan') if self.recall + self.precision == 0 else 2 * self.recall * self.precision / (self.recall + self.precision)

        self.miou = sum(self.ious) / len(self.ious) if len(self.ious) else 0
        self.map07 = len([x for x in self.ious if x >= 0.7]) / self.pd if self.pd != 0 else 0
        self.map08 = len([x for x in self.ious if x >= 0.8]) / self.pd if self.pd != 0 else 0
        self.map09 = len([x for x in self.ious if x >= 0.9]) / self.pd if self.pd != 0 else 0
        metric = dict(gt=self.gt, pd=self.pd, tp=self.tp, recall=self.recall, precision=[self.map07, self.map08, self.map09], f1=self.f1, miou=self.miou, idsw=self.id_switches, MOTA=self.MOTA)
        return metric


def compute_metric(data: Generator = None, IoU_thr: float = None, metric_types: List = None, save_path: str = None):
    """
    Computing IoU of all objects in all frames

    Args:
        data: Generator
            A generator object to get all information from all frames
        IoU_thr: float
            The iou threshold of tp boxes
        metric_types: List
            which type of box to compute metric, which can be chosen from ['2D'], ['3D'] and ['2D', '3D']
        save_path: str
            Local path to save metric results

    Returns:
        metric: List
            The metric of dataset which include two dict,
            the first represents metric of every single frame and the second represents metric of all frames

    Examples:
        .. code-block:: python

            from stardust.metric.tracking import compute_metric
            project_id = 856
            Export(project_id, 'top', input_path, True).export()
            json_datas = read_rosetta(project_id=project_id,
                                    input_path=input_path,
                                    )
            metric = compute_metric(json_datas, 0.5, ['2D', '3D'], 'local/')
    """
    assert data
    assert IoU_thr
    assert metric_types and metric_types in [['2D'], ['3D'], ['2D', '3D']]
    task_dict = collections.defaultdict(dict)

    for json_data in data:
        task_info = json_data.task_info
        task_id = str(task_info.task_id)
        frame_num = str(task_info.frame_num)
        task_dict[task_id][frame_num] = json_data
    metric_output = collections.defaultdict(dict)
    metric_total = {}
    for metric_type in metric_types:
        total_gt = total_pd = total_tp = total_idsw = 0
        if metric_type == '3D':
            for task_id in task_dict:
                task_info = sorted(task_dict[task_id].items(), key=lambda x: x[0])
                trajectory_gt = collections.defaultdict(list)
                trajectory_pd = collections.defaultdict(list)
                for frame_id in range(len(task_info)):
                    gts = task_info[frame_id][1].annotation.box3d_lst
                    pds = task_info[frame_id][1].prediction.box3d_lst
                    for obj_id, obj_info in gts.items():
                        x, y, z = obj_info.center
                        l, w, h = obj_info.size
                        ry = obj_info.rotation[-1]
                        t_data = TrackingData(obj_id, metric_type, [x, y, z, l, w, h, ry])
                        trajectory_gt[frame_id].append(t_data)

                    for obj_id, obj_info in pds.items():
                        x, y, z = obj_info.center
                        l, w, h = obj_info.size
                        ry = obj_info.rotation[-1]
                        t_data = TrackingData(obj_id, metric_type, [x, y, z, l, w, h, ry])
                        trajectory_pd[frame_id].append(t_data)

                metric_output[task_id][metric_type] = TrackingMetric(trajectory_gt, trajectory_pd, IoU_thr, metric_type).compute()
                total_gt += metric_output[task_id]['gt']
                total_pd += metric_output[task_id]['pd']
                total_tp += metric_output[task_id]['tp']
                total_idsw += metric_output[task_id]['idsw']

        if metric_type == '2D':
            for task_id in task_dict:
                task_info = sorted(task_dict[task_id].items(), key=lambda x: x[0])
                trajectory_gt = collections.defaultdict(list)
                trajectory_pd = collections.defaultdict(list)
                for frame_id in range(len(task_info)):
                    gts = task_info[frame_id][1].annotation.box2d_lst
                    pds = task_info[frame_id][1].prediction.box2d_lst
                    for obj_id, obj_info in gts.items():
                        x, y = obj_info.center
                        w, h = obj_info.size
                        t_data = TrackingData(obj_id, metric_type, [x, y, w, h])
                        trajectory_gt[frame_id].append(t_data)

                    for obj_id, obj_info in pds.items():
                        x, y = obj_info.center
                        w, h = obj_info.size
                        t_data = TrackingData(obj_id, metric_type, [x, y, w, h])
                        trajectory_pd[frame_id].append(t_data)

                metric_output[task_id] = TrackingMetric(trajectory_gt, trajectory_pd, IoU_thr, metric_type).compute()
                total_gt += metric_output[task_id]['gt']
                total_pd += metric_output[task_id]['pd']
                total_tp += metric_output[task_id]['tp']
                total_idsw += metric_output[task_id]['idsw']
        total_fn = total_gt - total_tp
        total_fp = total_pd - total_tp
        total_recall = float('nan') if total_gt == 0 else total_tp / total_gt
        total_precision = float('nan') if total_pd == 0 else total_tp / total_pd
        total_f1 = float('nan') if total_recall + total_precision == 0 else 2 * total_recall * total_precision / (total_recall + total_precision)
        total_MOTA = float('nan') if total_gt == 0 else 1 - (total_fn + total_fp + total_idsw) / float(total_gt)
        metric_total[metric_type] = dict(gt=total_gt, pd=total_pd, tp=total_tp, recall=total_recall, precision=total_precision, f1=total_f1, MOTA=total_MOTA, idsw=total_idsw)

    if save_path:
        os.makedirs(os.path.join(save_path, 'metric', 'tracking'), exist_ok=True)
        with open(os.path.join(save_path, 'metric', 'tracking', 'metric_by_task_id.json'), 'w') as f:
            json.dump(metric_output, f)
        with open(os.path.join(save_path, 'metric', 'tracking', 'metric_summary.json'), 'w') as f:
            json.dump(metric_total, f)
    return [metric_output, metric_total]



import numpy as np

from stardust.utils.frames import *
from stardust.utils.matrix4 import *


class SlotMapping:
    attr_map = {
        'point': 'point',
        'line': 'vertices',
        'splines': 'vertices',
        'polygon': 'vertices',
        'box2d': 'plane',
        'cuboid': 'vertices',
    }

    @classmethod
    def preposition(cls, data, father, frame_number, image_source):
        cls.target = frame_number
        cls.father = father
        cls.camera, cls.width, cls.height = image_source[data['source']].values()
        cls.time_series = data.get('timeSeries', {})
        if cls.time_series:
            cls.frame_list = np.array(list(cls.time_series.keys()), dtype=int).tolist()

    def __init__(self, data, father, frame_number, image_source):
        self.preposition(data, father, frame_number, image_source)
        for n in ['id', 'label', 'teamId', 'source', 'coordinate']:
            if data.get(n):
                setattr(self, n, data[n])
        self.type = data['type']
        commonality_result = self.commonality()
        if callable(commonality_result):
            if isinstance(commonality_result()(), bool):
                self.label = True
            setattr(self, self.attr_map[self.type], commonality_result()())
        else:
            if isinstance(commonality_result, bool):
                self.label = True
            setattr(self, self.attr_map[self.type], commonality_result)

    def commonality(self):
        if self.time_series and self.target in self.frame_list:
            if self.time_series[str(self.target)]['type'] == "disappear":
                return True
            else:
                input_value = self.time_series[str(self.target)].get("value")
                return input_value
        else:
            if isinstance(self.father.label, bool):
                return True
            else:
                return self.slot_factory

    def slot_factory(self):
        return {
            'point': self.point_mapping,
            'line': self.line_mapping,
            'splines': self.line_mapping,
            'polygon': self.line_mapping,
            'box2d': self.box2d_mapping,
            'cuboid': self.cuboid_mapping,
        }[self.type]

    def camera_indicator_info(self, pint_cloud_map_scope=None):
        """
            Camera coverage information
        """
        vx, vy, vz = self.camera.rotated_vector

        angle_x = np.arctan2(vy, vz) + (np.pi if vz < 0 else 0)
        angle_y = np.arctan2(vx, vz) + (np.pi if vy < 0 else 0)
        angle_z = np.arctan2(vy, vx) + (np.pi if vx < 0 else 0)

        adjustable_width = self.width * (pint_cloud_map_scope if pint_cloud_map_scope else 1)
        fx = self.camera.intrinsic[0]
        fx_with_width_proportion = fx / self.width
        _ = np.arctan(adjustable_width / 2 / fx / fx_with_width_proportion)

        result = {
            'position': self.camera.position,
            'rotation': dict(zip('xyz', (angle_x, angle_y, angle_z))),
            'thetaStart': np.pi / 2 - _,
            'thetaLength': 2 * _,
        }

        if self.camera.fov:
            theta_length = np.deg2rad(fov)
            result['thetaStart'] -= (theta_length - result['thetaLength']) / 2
            result['thetaLength'] = theta_length

        return result

    def point_in_camera(self, point):
        vector3d = np.array([point['x'], point['y'], point['z'], 1])
        return dict(zip('xyz', np.array(np.dot(self.camera.cam2lidar_rt.I, vector3d))[0, :3]))

    def pixel_coordinate(self, point):
        return dict(zip('xy', self.camera.get_pixel_coordinate(point)))

    # 2D point
    def point_mapping(self):
        father_data = self.father['point']
        son_data = self.point_in_camera(father_data)
        if son_data['z'] < 0 and (self.camera.fov is None or self.camera.fov < 180):
            return True
        return self.pixel_coordinate(son_data)

    def intersection(self, point1, point2, translation, plane1, plane2):
        v1, v2 = Vector(*point1).sub(translation), Vector(*point2).sub(translation)
        line = Line(v1, v2)
        direction = line.delta()
        result1 = intersect1 = plane1.intersect_line(line.clone(), direction)
        result2 = intersect2 = plane2.intersect_line(line.clone(), direction)
        if not intersect1 and not intersect2:
            return None
        if not intersect2 or result1.distance_to_squared(v1) < result2.distance_to_squared(v1):
            return self.point_in_camera(result1.add(translation))
        return self.point_in_camera(result2.add(translation))

    # 2D line, curve, or polygon
    def line_mapping(self):
        father_data = self.father['points']
        if not father_data:
            return True
        position, rotation, theta_start, theta_length = self.camera_indicator_info().values()
        plane = Plane(np.array([0, np.pi, 0]), 0)
        plane1 = plane.clone().apply_matrix4(make_rotation_z(theta_start + rotation['z']))
        plane2 = plane.clone().apply_matrix4(make_rotation_z(theta_start + theta_length + rotation['z']))
        filtered_points = [self.point_in_camera(p) for p in father_data if p['z'] != 0]  # Drop z == 0 points
        pairs = [(a, b) for a, b in zip(filtered_points, filtered_points[1:])]
        result = []
        for num, (pre, cur) in enumerate(pairs):
            pre_in, cur_in = check_point_in_camera(pre, theta_length), check_point_in_camera(cur, theta_length)
            if pre_in and cur_in:
                if num == 0:
                    result.append(self.pixel_coordinate(pre))
                result.append(self.pixel_coordinate(cur))
            elif not pre_in and cur_in:
                result.append(self.pixel_coordinate(self.intersection(pre, cur, Vector(*position), plane1, plane2)))
            elif pre_in and not cur_in:
                result.append(self.pixel_coordinate(self.intersection(pre, cur, Vector(*position), plane1, plane2)))
                break
        if not result:
            return True
        return result

    # 2D box
    def box2d_mapping(self):
        father_data = self.father['matrix']
        pass

    # 2D projection of a cuboid
    def cuboid_mapping(self):
        father_data = self.father['matrix']
        pass


if __name__ == '__main__':
    import numpy as np

    # Create a vector
    # vector = np.array([1, 2, 3, 1])  # Use a 4D vector (homogeneous coordinate) for 4x4 transforms
    #
    # # Create an identity matrix representing no transformation
    # matrix = np.matrix([[0.9995910729860709, 0.02835473731726321, 0.003700767248111604, 2.1812003147458467],
    #                     [-0.02843236707957861, 0.9993321887903348, 0.022951621942053253, -0.19653660649169938],
    #                     [-0.003047508623086898, -0.023047457976702046, 0.9997297271622989, -0.10265531758175112],
    #                     [0.0, 0.0, 0.0, 1.0]])
    #
    # transformed_vector = dict(zip('xyz', np.array(np.dot(matrix.I, vector))[0, :3]))
    # print(f"Transformed Vector: {transformed_vector}")
    # transformed_vector = np.dot(vector, matrix.I.T)
    # print(f"Transformed Vector: {transformed_vector}")
    # np.dot(matrix.I, vector) == np.dot(vector, matrix.I.T)
